/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <string.h>                     /* strcasecmp() */

#include "ansi_dec.h"
#include "avt0_dec.h"
#include "cons_buf.h"
#include "pctextmode.h"
#include "sauce.h"
#include "term_dyn.h"

#define ENABLE_BINARY
/* #define TEST_RENDER */

#define DEFAULT_BINARY_WIDTH 160

typedef enum{
    F_ANSI,
    F_AVATAR0,
    F_BINARY
}format_t;

struct _abydos_plugin_handle_t{
    format_t format;
    abydos_plugin_info_t *info;
    Console *console;
    Terminal *terminal;
    AnsiDecoder *ansi_dec;
    Avatar0Decoder *avt0_dec;
    int eof;
#ifdef ENABLE_BINARY
    int pos;
#endif
    sauce_buffer_t sauce_buffer;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
    int ice_color;
    int letter_spacing;
#ifndef TEST_RENDER
    cairo_surface_t *surface[2];
#endif
};



#ifdef TEST_RENDER
static void
_draw_screen(cairo_t *cr,Console *cons,ConsoleState state,int letter_spacing)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
    int x;
    int y;
    cairo_save(cr);
    for(y=0;y<height;++y){
        cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,80*letter_spacing,16);
        cairo_pattern_t* pattern = cairo_pattern_create_for_surface(surface);
        uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
        int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-letter_spacing;
        cairo_matrix_t matrix;
        for(x=0;x<width;++x){
            ConsoleCell cell = *src++;
            uint32_t bg;
            uint32_t fg;
            if(S_ICE==state){
                bg = _ega_color[(cell.attr>>4)&15];
            }else{
                bg = _ega_color[(cell.attr>>4)&7];
            }
            if(!(S_OFF==state && (cell.attr&0x80))){
                fg = _ega_color[(cell.attr)&15];
                _draw_character(dst+x*letter_spacing,rowskip,cell.ch,bg,fg,letter_spacing);
            }else{
                _draw_character(dst+x*letter_spacing,rowskip,' ',bg,bg,letter_spacing);
            }
        }
        cairo_surface_mark_dirty(surface);
        cairo_pattern_set_filter(pattern,CAIRO_FILTER_BEST);
        //cairo_pattern_set_filter(pattern,CAIRO_FILTER_BILINEAR);
        //cairo_pattern_set_filter(pattern,CAIRO_FILTER_NEAREST);
        cairo_pattern_get_matrix(pattern,&matrix);
        //matrix.yy *= 16.0/16.5;
        cairo_pattern_set_matrix(pattern,&matrix);
        cairo_pattern_set_extend(pattern,CAIRO_EXTEND_REPEAT);
        cairo_set_source(cr,pattern);
        cairo_save(cr);
        cairo_rectangle(cr,0,y*16,80*letter_spacing,16.6);
        cairo_clip(cr);
        cairo_paint(cr);
        cairo_restore(cr);
        //cairo_translate(cr,0,16);
        cairo_pattern_destroy(pattern);
        cairo_surface_destroy(surface);
    }
    cairo_restore(cr);
}
#else
#if 0
static cairo_surface_t *
_surface_from_console(Console *cons,ConsoleState state,int letter_spacing)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width*letter_spacing,height*16);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int rowskip = stride-letter_spacing;
    int x;
    int y;
    for(y=0;y<height;++y){
        for(x=0;x<width;++x){
            ConsoleCell cell = *src++;
            uint32_t bg;
            uint32_t fg;
            if(S_ICE==state){
                bg = abydos_cga_palette16[(cell.attr>>4)&15];
            }else{
                bg = abydos_cga_palette16[(cell.attr>>4)&7];
            }
            if(!(S_OFF==state && (cell.attr&0x80))){
                fg = abydos_cga_palette16[(cell.attr)&15];
                _draw_character(dst+x*letter_spacing,rowskip,cell.ch,bg,fg,letter_spacing);
            }else{
                _draw_character(dst+x*letter_spacing,rowskip,' ',bg,bg,letter_spacing);
            }
        }
        dst += stride*16;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif
#endif


static const char **
_ansi_get_mime_types()
{
    static const char *mime_types[] = {
        "text/x-ansi",
        "text/x-avatar0",
#ifdef ENABLE_BINARY
        "text/x-binary",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ansi_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    if(!strcasecmp(mime_type,"text/x-ansi")){
        h->format = F_ANSI;
    }else if(!strcasecmp(mime_type,"text/x-avatar0")){
        h->format = F_AVATAR0;
#ifdef ENABLE_BINARY
    }else if(!strcasecmp(mime_type,"text/x-binary")){
        h->format = F_BINARY;
        h->pos = 0;
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->console = cons_buf_new();
    cons_buf_set_size(h->console,F_BINARY==h->format ? DEFAULT_BINARY_WIDTH : 80,1,7);
    h->terminal = (F_ANSI==h->format || F_AVATAR0==h->format) ? term_dyn_new(&cons_buf_if,h->console) : NULL;
    h->ansi_dec = (F_ANSI==h->format || F_AVATAR0==h->format) ? ansi_decoder_new(&term_dyn_if,h->terminal,NULL,NULL) : NULL;
    h->avt0_dec = (F_AVATAR0==h->format) ? avatar0_decoder_new(&term_dyn_if,h->terminal) : NULL;
    h->eof = 0;
    sauce_buffer_init(&h->sauce_buffer);
#ifndef TEST_RENDER
    h->surface[0] = NULL;
#endif
    return h;
}


static void
_ansi_free(abydos_plugin_handle_t *h)
{
    if(h->ansi_dec)
        ansi_decoder_free(h->ansi_dec);
    if(h->avt0_dec)
        avatar0_decoder_free(h->avt0_dec);
    if(h->terminal)
        term_dyn_free(h->terminal);
    cons_buf_free(h->console);
#ifndef TEST_RENDER
    if(h->surface[0]){
        cairo_surface_destroy(h->surface[0]);
        if(h->info->frame_count>1)
            cairo_surface_destroy(h->surface[1]);
    }
#endif
    free(h);
}



static void
_ansi_progressive_begin(abydos_plugin_handle_t *h,abydos_plugin_info_func_t info_func,abydos_plugin_progress_func_t progress_func,void *userdata)
{
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
}


static int
_ansi_progressive_feed(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    while(len>0){
        int ch = (unsigned char)*data++;
        --len;
        if(h->eof){
            sauce_buffer_add(&h->sauce_buffer,ch);
        }else{
            switch(h->format){
            case F_AVATAR0:
                if(avatar0_decoder_write(h->avt0_dec,ch))
                    continue;
                /* fall through */
            case F_ANSI:
                if(ansi_decoder_write(h->ansi_dec,ch))
                    continue;
                if(26==ch){
                    h->eof = 1;
                    continue;
                }
                term_dyn_write(h->terminal,ch);
                break;
#ifdef ENABLE_BINARY
            case F_BINARY:
                if(h->sauce_buffer.size&1){
                    int ch0 = h->sauce_buffer.buffer[h->sauce_buffer.size-1];
                    int x = h->pos % cons_buf_width(h->console);
                    int y = h->pos / cons_buf_width(h->console);
                    if(0==x)
                        cons_buf_extend(h->console,1,0x07);
                    cons_buf_charxya(h->console,x,y,ch,ch0);
                    ++h->pos;
                }
                sauce_buffer_add(&h->sauce_buffer,ch);
                if(h->sauce_buffer.size >= SAUCE_BUFFER_SIZE){
                    memcpy(h->sauce_buffer.buffer,h->sauce_buffer.buffer+SAUCE_BUFFER_SIZE-128,128);
                    h->sauce_buffer.size = 128;
                }
#endif
            }
        }
    }
    return 0;
}


static int
_ansi_progressive_end(abydos_plugin_handle_t *h)
{
    sauce_t sauce;
    if(sauce_buffer_validate(&h->sauce_buffer,&sauce)){
        DEBUG_p("found valid sauce tag");
        h->ice_color = sauce_ANSiFlags_iCE_Color(&sauce);
        h->letter_spacing = sauce_ANSiFlags_Letter_Spacing(&sauce);
        if(sauce_ANSiFlags_Aspect_Ratio(&sauce)){
            h->info->pixel_ratio = 1;
        }else{
            h->info->pixel_ratio = 20.0/(3*h->letter_spacing);
        }
#ifdef ENABLE_BINARY
        if(F_BINARY==h->format){
            int width = sauce_BinaryText_width(&sauce);
            int height = (sauce.FileSize/2)/width;
            DEBUG_p("width: %d",width);
            DEBUG_p("filesize: %d",(int)sauce.FileSize);
            cons_buf_reshape(h->console,width);
            cons_buf_set_size(h->console,width,height,7);
        }
#endif
    }else{
        DEBUG_p("no valid sauce tag found");
        h->ice_color = 0;
        h->letter_spacing = 9;
        h->info->pixel_ratio = 20.0/27;
    }
    h->info->width = cons_buf_width(h->console)*h->letter_spacing;
    h->info->height = cons_buf_height(h->console)*16;
    h->info->threadsafe = 1;
#ifndef TEST_RENDER
    h->surface[0] = abydos_image_surface_from_console(h->console,h->ice_color ? S_ICE : S_ON,h->letter_spacing);
#endif
    if(!h->ice_color && abydos_console_has_blink(h->console)){
        h->info->frame_count = 2;
#ifndef TEST_RENDER
        h->surface[1] = abydos_image_surface_from_console(h->console,S_OFF,h->letter_spacing);
#endif
    }

    if(h->info_func)
        h->info_func(h->userdata);
    if(h->progress_func){
        cairo_rectangle_int_t r;
        r.x = 0;
        r.y = 0;
        r.width = h->info->width;
        r.height = h->info->height;
        h->progress_func(h->userdata,&r);
    }
    return 0;
}


#ifdef TEST_RENDER
static void
_ansi_render(abydos_plugin_handle_t *h, cairo_t *cr,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    _draw_screen(cr,h->console,h->ice_color ? S_ICE : (frame&1 ? S_OFF : S_ON),h->letter_spacing);
}
#else
static cairo_surface_t *
_ansi_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface[frame]);
}
#endif


static void
_ansi_get_duration(abydos_plugin_handle_t *h,int page,double *frame_duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        frame_duration[i] = BLINK_DURATION;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ansi_get_mime_types,
        .new = _ansi_new,
        .free = _ansi_free,
        .progressive_begin = _ansi_progressive_begin,
        .progressive_feed = _ansi_progressive_feed,
        .progressive_end = _ansi_progressive_end,
#ifdef TEST_RENDER
        .render = _ansi_render,
#else
        .get_image_surface = _ansi_get_image_surface,
#endif
        .get_duration = _ansi_get_duration
    };
    return &p;
}
