/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <png.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_apng_get_mime_types()
{
    static const char *mime_types[] = {
        "image/apng",
        "image/png",
        "image/vnd.mozilla.apng",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_apng_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_apng_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_rows(abydos_plugin_handle_t *h, png_bytepp rows)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    for(y=0;y<h->info->height;++y){
        png_bytep row = rows[y];
#if 1
        int x;
        for(x=0;x<h->info->width;++x){
            int alpha = row[x*4+3];
            dst[x] = ((alpha*row[x*4+0]/255)<<16)|
                     ((alpha*row[x*4+1]/255)<< 8)|
                     ((alpha*row[x*4+2]/255)<< 0)|
                     (alpha<<24);
        }
#else
        nile_copy_colors_rgb(dst, 4, NIL_CAIRO_FORMAT_ARGB32, row, 4, NILE_FORMAT_r8g8b8a8, h->info->width);
#endif
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_apng_create_from_stream(abydos_plugin_handle_t *h,FILE *f)
{
    png_structp png;
    png_infop info;
    double gamma;
    png_uint_32 width;
    png_uint_32 height;
    png_uint_32 rowbytes;
    png_bytepp rows;
    int i;

    png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(!png){
        return -1;
    }
    info = png_create_info_struct(png);
    if(!info){
        png_destroy_read_struct(&png, NULL, NULL);
        return -1;
    }

    png_init_io(png,f);
    png_read_info(png,info);
    png_set_expand(png);
    png_set_strip_16(png);
    png_set_gray_to_rgb(png);
    png_set_add_alpha(png,0xff,PNG_FILLER_AFTER);
    //png_set_bgr(png);
    png_set_interlace_handling(png);
    if(png_get_gAMA(png, info, &gamma)){
        png_set_gamma(png, PNG_DEFAULT_sRGB, gamma);
    }
    png_read_update_info(png,info);
    width = png_get_image_width(png,info);
    height = png_get_image_height(png,info);
    rowbytes = png_get_rowbytes(png,info);
    h->info->width = width;
    h->info->height = height;

    if(png_get_valid(png,info,PNG_INFO_pHYs)){
        unsigned int num = 1;
        unsigned int den = 1;
        int units = 0;
        png_get_pHYs(png,info,&den,&num,&units);
        if(num > 0 && den > 0)
            h->info->pixel_ratio = (double)num / den;
        DEBUG_p("pixel ratio: %f (%d/%d)",h->info->pixel_ratio,num,den);
    }



#ifdef PNG_APNG_SUPPORTED
    if(png_get_valid(png,info,PNG_INFO_acTL)){
        png_uint_32 frames;
        png_uint_32 plays = 0;
        png_get_acTL(png,info,&frames,&plays);
        h->info->frame_count = frames;
    }
#endif
    h->frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));

    rows = (png_bytepp)malloc(height*sizeof(png_bytep));
    rows[0] = (uint8_t*)malloc(height*rowbytes);
    for(i=1;i<height;++i)
        rows[i] = rows[0] + i*rowbytes;

    for(i=0;i<h->info->frame_count;++i){
        h->frame[i].duration = 0;
#ifdef PNG_APNG_SUPPORTED
        png_uint_32 xoffs;
        png_uint_32 yoffs;
        if(png_get_valid(png,info,PNG_INFO_acTL)){
            uint16_t delay_num = 0;
            uint16_t delay_den;
            uint8_t dop;
            uint8_t bop;
            png_read_frame_head(png,info);
            png_get_next_frame_fcTL(png,info,&width,&height,&xoffs,&yoffs,&delay_num,&delay_den,&dop,&bop);
            DEBUG_p("daley_den: %d",(int)delay_den);
            h->frame[i].duration = (double)delay_num/delay_den;
        }
#endif
        png_read_image(png,rows);
        h->frame[i].surface = _surface_from_rows(h,rows);
    }
    free(rows[0]);
    free(rows);
    png_destroy_read_struct(&png, &info, NULL);
    return 0;
}


static int
_apng_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    FILE *f;
    int res;
    f = fmemopen((void*)bytes,len,"r");
    if(!f){
        return -1;
    }
    res = _apng_create_from_stream(h,f);
    fclose(f);
    return res;
}


static int
_apng_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    FILE *f;
    int res;
    f = fopen(filename,"r");
    if(!f)
        return -1;
    res = _apng_create_from_stream(h,f);
    fclose(f);
    return res;
}


static cairo_surface_t *
_apng_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_apng_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _apng_get_mime_types,
        .new = _apng_new,
        .free = _apng_free,
        .create_from_data = _apng_create_from_data,
        .create_from_file = _apng_create_from_file,
        .get_image_surface = _apng_get_image_surface,
        .get_duration = _apng_get_duration
    };
    return &p;
}
