/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "atarist.h"


#define DEFAULT_DURATION (5.0/60)


typedef struct{
    cairo_surface_t *surface;
}frame_t;

typedef struct{
    frame_t *frame;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_MBK,
        T_PAC
    }type;
    int resolution;
    page_t *page;
};


static const char **
_stos_get_mime_types()
{
    static const char *mime_types[] = {
        "application/x-stos-memorybank",
        "image/x-stos-picturepacker",
        "image/x-stos-picturepacker-lowres",
        "image/x-stos-picturepacker-medres",
        "image/x-stos-picturepacker-hires",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_stos_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    if(!strncasecmp(mime_type,"image/x-stos-picturepacker",26)){
        h->type = T_PAC;
        if(!strncasecmp(mime_type+26,"-l",2)){
            h->resolution = 0;
        }else if(!strncasecmp(mime_type+26,"-m",2)){
            h->resolution = 1;
        }else if(!strncasecmp(mime_type+26,"-h",2)){
            h->resolution = 2;
        }else{
            h->resolution = -1;
        }
    }else{
        h->type = T_MBK;
        h->resolution = -1;
    }
    h->info = info;
    info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_stos_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int j;
        for(j=0;j<h->info->page_count;++j){
            int i;
            for(i=0;i<h->info->frame_count;++i)
                cairo_surface_destroy(h->page[j].frame[i].surface);
            free(h->page[j].frame);
        }
        free(h->page);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get24(a) (get16(a)<<8|get8(a+2))
#define get32(a) be32toh(*(uint32_t*)(a))


static int
_decompress(uint8_t *dst,
            int len,
            const uint8_t *table1,
            const uint8_t *table2,
            const uint8_t *pointers)
{
    int i;
    for(i=0;i<len;++i){
        if(!(i&7))
            if(pointers[i>>6] & (128>>((i>>3)&7)))
                ++table2;
        if(*table2 & (128>>(i&7)))
            ++table1;
        *dst++ = *table1;
    }
    return 0;
}


cairo_surface_t *
_surface_from_buffer_0(const uint8_t *src,
                        int tx,
                        int ty,
                        int tcar,
                        atarist_cairo_color_t *cairo_palette)
{
    int width = tx*16;
    int height = ty*tcar;
    int planesize = tx*2*ty*tcar;
    cairo_surface_t *surface;
    atarist_cairo_color_t* dst;
    int stride;
#ifdef ATARIST_CAIRO_BITS16
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
#else
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
#endif
    dst = (atarist_cairo_color_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(atarist_cairo_color_t);
    while(ty-->0){
        int x;
        for(x=0;x<tx*2;++x){
            atarist_cairo_color_t *dst2 = dst + x*8;
            int y;
            for(y=0;y<tcar;++y){
                int plane1 = src[planesize*0];
                int plane2 = src[planesize*1];
                int plane3 = src[planesize*2];
                int plane4 = src[planesize*3];
                int i;
                for(i=0;i<8;++i){
                    int pixel = (plane1 & (128>>i) ? 0x01 : 0) |
                                (plane2 & (128>>i) ? 0x02 : 0) |
                                (plane3 & (128>>i) ? 0x04 : 0) |
                                (plane4 & (128>>i) ? 0x08 : 0);
                    *dst2++ = cairo_palette[pixel];
                }
                ++src;
                dst2 += stride-8;
            }
        }
        dst += tcar*stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


/*
 * The graphics memory of the Atari ST is 32KB, separated into bitplanes
 * depending on the resolution. For low resolution therse are four layers
 * (8KB each), for medium resolution there are two (16KB each) and for high
 * resolution these is only one. The way they are aranged in memory is one
 * word (16bits) from the first bitplane, then one word from the second
 * bitplane and so on. This matters because the compression algorithm first
 * compresses all the bits of the first layer, then all bits of the second
 * layer and so of (presumably because this achieves better compression than
 * just compressing the whole grapahics memory as is from start to finish).
 * In short this means that the uncompressed bits can be in three different
 * orders depending on the (intended) resolution.
 *
 * However some files have the wrong resolution set, and have also been
 * compressed for that resolution. This yealds the need for this special
 * functions that that shuffels the bits around as for medium resultion but
 * interpretes them as a low resolution image.
 */
cairo_surface_t *
_surface_from_buffer_0_as_1(const uint8_t *src,
                        int tx,
                        int ty,
                        int tcar,
                        atarist_cairo_color_t *cairo_palette)
{
    int width = tx*8;
    int height = ty*tcar;
    int planesize = tx*2*ty*tcar;
    cairo_surface_t *surface;
    atarist_cairo_color_t* dst;
    int stride;
#ifdef ATARIST_CAIRO_BITS16
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
#else
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
#endif
    dst = (atarist_cairo_color_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(atarist_cairo_color_t);
    while(ty-->0){
        int x;
        for(x=0;x<tx/2;++x){
            int byte;
            for(byte=0;byte<2;++byte){
                int y;
                for(y=0;y<tcar;++y){
                    int plane1 = src[planesize*0 + tcar*0];
                    int plane2 = src[planesize*1 + tcar*0];
                    int plane3 = src[planesize*0 + tcar*2];
                    int plane4 = src[planesize*1 + tcar*2];
                    int i;
                    for(i=0;i<8;++i){
                        int pixel = (plane1 & (128>>i) ? 0x01 : 0) |
                                    (plane2 & (128>>i) ? 0x02 : 0) |
                                    (plane3 & (128>>i) ? 0x04 : 0) |
                                    (plane4 & (128>>i) ? 0x08 : 0);
                        dst[x*16 + byte*8 + i + y*stride] = cairo_palette[pixel];
                    }
                    ++src;
                }
            }
            src += tcar*2;
        }
        dst += tcar*stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
_surface_from_buffer_1(const uint8_t *src,
                       int tx,
                       int ty,
                       int tcar,
                       atarist_cairo_color_t *cairo_palette)
{
    int width = tx*16;
    int height = ty*tcar;
    int planesize = tx*2*ty*tcar;
    cairo_surface_t *surface;
    atarist_cairo_color_t* dst;
    int stride;
#ifdef ATARIST_CAIRO_BITS16
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
#else
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
#endif
    dst = (atarist_cairo_color_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(atarist_cairo_color_t);
    while(ty-->0){
        int x;
        for(x=0;x<tx*2;++x){
            atarist_cairo_color_t *dst2 = dst + x*8;
            int y;
            for(y=0;y<tcar;++y){
                int plane1 = src[planesize*0];
                int plane2 = src[planesize*1];
                int i;
                for(i=0;i<8;++i){
                    int pixel = (plane1 & (128>>i) ? 0x01 : 0) |
                                (plane2 & (128>>i) ? 0x02 : 0);
                    *dst2++ = cairo_palette[pixel];
                }
                ++src;
                dst2 += stride-8;
            }
        }
        dst += tcar*stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


/*
 * This function is untested.
 */
cairo_surface_t *
_surface_from_buffer_2(const uint8_t *src,
                       int tx,
                       int ty,
                       int tcar)
{
    int width = tx*16;
    int height = ty*tcar;
    cairo_surface_t *surface;
    atarist_cairo_color_t* dst;
    int stride;
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    dst = (uint16_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(uint16_t);
    while(ty-->0){
        int x;
        for(x=0;x<tx*2;++x){
            atarist_cairo_color_t *dst2 = dst + x*8;
            int y;
            for(y=0;y<tcar;++y){
                int plane1 = src[0];
                int i;
                for(i=0;i<8;++i)
                    *dst++ = (plane1 & (128>>i)) ? 0xffff : 0x0000;
                ++src;
                dst2 += stride-8;
            }
        }
        dst += tcar*stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
_surface_from_buffer_2_as_1(const uint8_t *src,
                            int tx,
                            int ty,
                            int tcar)
{
    int width = tx*16;
    int height = ty*tcar*2;
    int planesize = tx*2*ty*tcar;
    cairo_surface_t *surface;
    atarist_cairo_color_t* dst;
    int stride;
    DEBUG_p("dimensions: %dx%d",width,height);
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    dst = (uint16_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(uint16_t);
    while(ty-->0){
        int x;
        for(x=0;x<tx;++x){
            int y;
            for(y=0;y<tcar;++y){
                uint32_t bits;
                int i;
                bits = src[planesize*0 + tcar*0] << 24 |
                       src[planesize*0 + tcar*1] << 16 |
                       src[planesize*1 + tcar*0] << 8 |
                       src[planesize*1 + tcar*1];
                for(i=0;i<32;++i)
                    dst[x*32 + i + y*stride*2] = bits & (1<<(31-i)) ? 0x0000 : 0xffff;
                ++src;
            }
            src += tcar;
        }
        dst += tcar*stride*2;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_handle_packed_screen(abydos_plugin_handle_t *h,const uint8_t *data,size_t len)
{
    atarist_cairo_color_t palette[16];
    int resolution = get16(data+0);
    int dx = get16(data+2);
    int dy = get16(data+4);
    int tx = get16(data+6);
    int ty = get16(data+8);
    int unknown0 = get16(data+10);
    int tcar = get16(data+12);
    int flags = get16(data+14);
    int as_1_hack = 0;
    uint32_t table1 = 70-4;
    uint32_t table2 = get32(data+16)-4;
    uint32_t pointers = get32(data+20)-4;
    /* 24 - 55: palette */
    size_t buffer_len = tx*8*ty*tcar;
    uint8_t *buffer;
    int res;

/*

    Header of file created by STOS:

    resulution: 0
    dx: 0
    dy: 0
    tx: 20
    ty (height): 5
    tcar (blocksize): 40
    unknown0: 0
    flags: 2
    table1: 66 (42)
    table2: 10889 (2a89)
    pointers: 13299 (33f3)
    end: 13801 (35e9)

*/

    DEBUG_p("resulution: %d",resolution);
    DEBUG_p("dx: %d",dx);
    DEBUG_p("dy: %d",dy);
    DEBUG_p("tx: %d",tx);
    DEBUG_p("ty (height): %d",ty);
    DEBUG_p("unknown0: %d",unknown0);
    DEBUG_p("tcar (blocksize): %d",tcar);
    DEBUG_p("flags: %d",flags);

    DEBUG_p("table1: %d (%x)",table1,table1);
    DEBUG_p("table2: %d (%x)",table2,table2);
    DEBUG_p("pointers: %d (%x)",pointers,pointers);
    DEBUG_p("end: %d (%x)",(int)len,(int)len);

    DEBUG_p("dimensions: %dx%d",tx*16,ty*tcar);
    DEBUG_p("buffer size: %d",(int)buffer_len);

    /* FIXME!! sanity check before allocating buffer */
    if(buffer_len <= 0){
        DEBUG_p("Empty data");
        h->info->error = "Empty data";
        return -1;
    }

#if 1
    /* Force resolution */
    switch(h->resolution){
    case 0:
        if(1==resolution){
            resolution = 0;
            as_1_hack = 1;
        }else if(2==resolution){
            /* Can this occur? */
            DEBUG_p("Image compressed as hires, but intended for lowres");
            return -1;
        }
        break;
    case 1:
        if(resolution!=1){
            /* Can this occur? */
            DEBUG_p("Image compressed as lowres or hires, but intended for medres");
            return -1;
        }
        break;
    case 2:
        if(0==resolution){
            /* Can this occur? */
            DEBUG_p("Image compressed as lowres, but intended for hires");
            return -1;
        }else if(1==resolution){
            resolution = 2;
            as_1_hack = 1;
        }
        break;
    }
#else
    /* Try to detect images that claim to be medium resolution but are
     * actually intended for medium resolution.
     */
    if(1==resolution && flags & 1){
        resolution = 0;
        as_1_hack = 1;
    }
#endif

    buffer = (uint8_t*)malloc(buffer_len);
    res = _decompress(buffer,
                      tx*8*ty*tcar,
                      data+table1,
                      data+table2,
                      data+pointers);

    if(res >= 0){
        cairo_surface_t *surface;
        h->page = (page_t*)malloc(sizeof(page_t));
        h->page[0].frame = (frame_t*)malloc(sizeof(frame_t));
        switch(resolution){
        case 0:
            atarist_decode_palette((uint16_t*)(data+34),palette);
            if(as_1_hack){
                surface = _surface_from_buffer_0_as_1(buffer, tx, ty, tcar, palette);
            }else{
                surface = _surface_from_buffer_0(buffer, tx, ty, tcar, palette);
            }
            h->info->width = 320;
            h->info->height = 200;
            h->info->pixel_ratio = 5.0/6;
            break;
        case 1:
            atarist_decode_palette((uint16_t*)(data+34),palette);
            surface = _surface_from_buffer_1(buffer, tx, ty, tcar, palette);
            h->info->width = 640;
            h->info->height = 200;
            h->info->pixel_ratio = 5.0/12;
            break;
        case 2:
            if(as_1_hack){
                surface = _surface_from_buffer_2_as_1(buffer, tx, ty, tcar);
            }else{
                surface = _surface_from_buffer_2(buffer, tx, ty, tcar);
            }
            h->info->width = 640;
            h->info->height = 400;
            h->info->pixel_ratio = 5.0/6;
            break;
        }
        h->page[0].frame[0].surface = surface;
    }
    free(buffer);
    return res;
}


cairo_surface_t *
_surface_from_frame_0(const uint8_t *src,
                      const uint8_t *mask,
                      int word_width,
                      int height,
                      atarist_cairo_color_t *cairo_palette)
{
    int planesize = 2 /* *word_width */;
    cairo_surface_t *surface;
    atarist_cairo_color_t* dst;
    int stride;
#ifdef ATARIST_CAIRO_BITS16
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,word_width*16,height);
#else
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,word_width*16,height);
#endif
    dst = (atarist_cairo_color_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(atarist_cairo_color_t);
    while(height-->0){
        int x;
        for(x=0;x<word_width;++x){
            int byte;
            for(byte=0;byte<2;++byte){
                int plane1 = src[byte+planesize*0];
                int plane2 = src[byte+planesize*1];
                int plane3 = src[byte+planesize*2];
                int plane4 = src[byte+planesize*3];
                int i;
                for(i=0;i<8;++i){
                    int pixel = (plane1 & (128>>i) ? 0x01 : 0) |
                                (plane2 & (128>>i) ? 0x02 : 0) |
                                (plane3 & (128>>i) ? 0x04 : 0) |
                                (plane4 & (128>>i) ? 0x08 : 0);
                    dst[x*16+byte*8+i] = cairo_palette[pixel] | 0xff<<24;
                }
            }
            src += 2*4;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_handle_sprites(abydos_plugin_handle_t *h,const uint8_t *data,size_t len)
{
    atarist_cairo_color_t palette[16];
    int offset = get32(data);
    int i;
    h->info->frame_count = get16(data+12);
    data += offset;
    len -= offset;
    atarist_decode_palette((uint16_t*)(data+h->info->frame_count*8 + 4), palette);
    h->page = (page_t*)malloc(sizeof(page_t));
    h->page[0].frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        const uint8_t *p = data + i*8;
        int width = get8(p+4);
        int height = get8(p+5);
        //int hx = get8(p+6);
        //int hy = get8(p+7);
        const uint8_t *mask;
        const uint8_t *color;
        offset = get32(p);
        DEBUG_p("frame %d offset: %d",i,offset);
        mask = data + offset;
        color = mask + 4*height;
        h->page[0].frame[i].surface = _surface_from_frame_0(color,mask,width,height,palette);
        h->info->width = width*16;
        h->info->height = height;
    }
    h->info->pixel_ratio = 5.0/6;
    return 0;
}


static cairo_surface_t *
_surface_from_icon(uint32_t bg,uint32_t fg,const uint16_t *src){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,16,16);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    for(y=0;y<16;++y){
        int mask = be16toh(*src++);
        int data = be16toh(*src++);
        int x;
        for(x=0;x<16;++x){
            if(mask&(1<<15)){
                dst[x] = (data&(1<<15) ? fg : bg) | 255<<24;
            }else{
                dst[x] = 0;
            }
            mask <<= 1;
            data <<= 1;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_handle_icons(abydos_plugin_handle_t *h,const uint8_t *data,size_t len)
{
    int i;
    h->info->page_count = get16(data+0);
    data += 2;
    len -= 2;
    if(h->info->page_count*84 > len)
        return -1;
    h->info->width = 16;
    h->info->height = 16;
    h->page = (page_t*)malloc(h->info->page_count*sizeof(page_t));
    for(i=0;i<h->info->page_count;++i){
        uint32_t bg = get16(data+6) ? 0xffffff : 0x000000;
        uint32_t fg = get16(data+8) ? 0xffffff : 0x000000;
        DEBUG_p("page %d - form: %d, bg: %d, fg: %d",i,get16(data+4),get16(data+6),get16(data+8));
        h->page[i].frame = (frame_t*)malloc(sizeof(frame_t*));
        h->page[i].frame[0].surface = _surface_from_icon(bg,fg,(const uint16_t*)(data+10));
        data += 84;
        len -= 84;
    }
    return 0;
}


static int
_handle_data(abydos_plugin_handle_t *h,const uint8_t *data,size_t len)
{
    uint32_t datatype = get32(data);
    DEBUG_p("data type: %x",datatype);
    data += 4;
    len -= 4;
    switch(datatype){
    case 0x003d2067: /* blocks */
        /* not handled yet */
        return -1;
    case 0x06071963: /* packed screen */
        return _handle_packed_screen(h,data,len);
    case 0x19861987: /* sprites */
        return _handle_sprites(h,data,len);
    case 0x28091960: /* icons */
        return _handle_icons(h,data,len);
    case 0x43410100: /* packed crack art file */
        /* not handled yet */
        return -1;
    case 0x544e5920: /* TINY pack screen */
        /* not handled yet */
        return -1;
    }
    h->info->error = "Invalid file";
    return -1;
}


static int
_handle_memorybank(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    //int banknum;
    int banktype;
    int size;
    if(strncmp(data,"Lionpoubnk",10)){
        h->info->error = "Invalid header";
        return -1;
    }
    //banknum = get32(data+10);
    banktype = get8(data+14);
    size = get24(data+15);
    switch(banktype){
    case 0x01: /* work */
    case 0x02: /* screen */
        return -1;
    case 0x81: /* data */
        return _handle_data(h,(const uint8_t*)data+18,size);
    case 0x82: /* datascreen */
    case 0x83: /* executable subroutines? */
    case 0x84: /* set */
    case 0x85: /* packed files */
        return -1;
    }
    return -1;
}


static int
_stos_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    switch(h->type){
    case T_MBK:return _handle_memorybank(h,data,len);
    case T_PAC:return _handle_data(h,(const uint8_t*)data,len);
    }
    return -1;
}


static cairo_surface_t *
_stos_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->page[page].frame[frame].surface);
}


static void
_stos_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = DEFAULT_DURATION;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _stos_get_mime_types,
        .new = _stos_new,
        .free = _stos_free,
        .create_from_data = _stos_create_from_data,
        .get_image_surface = _stos_get_image_surface,
        .get_duration = _stos_get_duration
    };
    return &p;
}
