/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>

#include "cgmdraw-cairo.h"


#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif


/* Arbitrary limit to prevent creation of huge bitmaps. */
#define MAXWIDTH 4096
#define MAXHEIGHT 4096


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_cgm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/cgm",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_cgm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_cgm_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_t *
_on_page(void *userdata, int width, int height)
{
    abydos_plugin_handle_t *h = userdata;
    double scale;
    cairo_rectangle_t ext;
    cairo_t *cr;
    DEBUG_p("begin");
    scale = MIN(1, MIN((double)MAXWIDTH/width, (double)MAXHEIGHT/height));
    ext.x = 0;
    ext.y = 0;
    ext.width = h->info->width = scale * width;
    ext.height = h->info->height = scale * height;
    h->surface = cairo_recording_surface_create(CAIRO_CONTENT_COLOR_ALPHA, &ext);
    cr = cairo_create(h->surface);
    cairo_scale(cr, scale, scale);
    return cr;
}


static int
_cgm_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    return cgm_draw_cairo((uint8_t*)data, len, _on_page, h);
}


static void
_cgm_render(abydos_plugin_handle_t *h,
            cairo_t *cr,
            int page,
            cairo_bool_t *layer_visibility,
            int variant,
            double time,
            int frame)
{
    cairo_save(cr);
    cairo_set_source_surface(cr, h->surface, 0.0, 0.0);
    cairo_paint(cr);
    cairo_restore(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _cgm_get_mime_types,
        .new = _cgm_new,
        .free = _cgm_free,
        .create_from_data = _cgm_create_from_data,
        .render = _cgm_render
    };
    return &p;
}
