/*
 * Copyright (C) 2000 – 2020  LibreOffice contributors
 * Copyright (C) 2021         Magnus Bergman <magnus@snisurset.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * This file incorporates work covered by the following license notice:
 *
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements. See the NOTICE file distributed
 *   with this work for additional information regarding copyright
 *   ownership. The ASF licenses this file to you under the Apache
 *   License, Version 2.0 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.apache.org/licenses/LICENSE-2.0 .
 */
#include "cgmparse.h"
#include "cgmread.h"
#include "elements.h"


int handle_class2(cgm_context_t *c, int id, cgm_read_t *r)
{
    switch(id){
    case 0x01 : /* Scaling Mode (E, R)*/
        if ( !cgm_read_eoe(r) ){    // HACK (NASA.CGM)
            c->pElement.eScalingMode = r_get_scaling_mode(r);
            c->pElement.nScalingFactor = r_get_R(r);
            cgm_set_map_mode(c, r);
        }
        break;
    case 0x02 : /* Colour Selection Mode (E) */
        c->pElement.eColorSelectionMode = r_get_color_selection_mode(r);
        break;
    case 0x03 : /* Line Width Specification Mode (E) */
        c->pElement.attr.eLineWidthSpecMode = r_get_spec_mode(r);
        break;
    case 0x04 : /* Marker Size Specification Mode (E) */
        c->pElement.attr.eMarkerSizeSpecMode = r_get_spec_mode(r);
        break;
    case 0x05 : /* Edge Width Specification Mode (E) */
        c->pElement.attr.eEdgeWidthSpecMode = r_get_spec_mode(r);
        break;
    case 0x06 : /* VDC Extent (P, P) */
        {
            loa_point_t p1;
            loa_point_t p2;
            r_get_P(r, &p1);
            r_get_P(r, &p2);
            c->pElement.aVDCExtent.pos.x = p1.x;
            c->pElement.aVDCExtent.pos.y = p2.y;
            c->pElement.aVDCExtent.size.x = p2.x - p1.x;
            c->pElement.aVDCExtent.size.y = p1.y - p2.y;
            cgm_set_map_mode(c, r);
        }
        break;
    case 0x07 : /* Background Color (CD) */
        c->pElement.nBackGroundColor = r_get_color(r, &c->pElement, 1);
        break;
    case 0x08 : /* Device Viewport (VP, VP)*/
        if ( r->vdc_type == VDC_INT )
            r_get_rect(r, &c->pElement.aDeviceViewPort );
        cgm_set_map_mode(c, r);
        break;
    case 0x09 : /* Device Viewport Specification Mode (VC, R) */
        r->eDeviceViewPortMode = r_get_device_view_port_mode(r);
        c->pElement.nDeviceViewPortScale = r_get_R(r);
        cgm_set_map_mode(c, r);
        break;
    case 0x0a : /* Device Viewport Mapping (E, E, E) */
        c->pElement.eDeviceViewPortMap = r_get_device_view_port_map(r);
        c->pElement.eDeviceViewPortMapH = r_get_device_view_port_map_h(r);
        c->pElement.eDeviceViewPortMapV = r_get_device_view_port_map_v(r);
        cgm_set_map_mode(c, r);
        break;
    case 0x0b : /* Line Representation (IX, IX, SS, CO) */
        {
            int idx = r_get_IX(r);
            LineBundle *bn = (LineBundle*)cgm_get_bundle(&c->pElement.attr.line_list, idx);
            if(!bn){
                bn = loa_array_append(&c->pElement.attr.line_list, 1);
                bn->base.idx = idx;
            }
            bn->type = (line_type_t)r_get_IX(r);
            bn->width = r_get_V(r, c->pElement.attr.eLineWidthSpecMode);
            bn->base.color = r_get_color(r, &c->pElement, 0);
            c->pElement.attr.line_bundled = loa_array_index(&c->pElement.attr.line_list, bn);
        }
        break;
    case 0x0c : /* Marker Representation (IX, IX, SS, CO) */
        {
            int idx = r_get_IX(r);
            MarkerBundle *bn = (MarkerBundle*)cgm_get_bundle(&c->pElement.attr.marker_list, idx);
            if(!bn){
                bn = loa_array_append(&c->pElement.attr.marker_list, 1);
                bn->base.idx = idx;
            }
            bn->eMarkerType = (MarkerType)r_get_IX(r);
            bn->nMarkerSize = r_get_V(r, c->pElement.attr.eMarkerSizeSpecMode);
            bn->base.color = r_get_color(r, &c->pElement, 0);
            c->pElement.attr.marker_bundled = loa_array_index(&c->pElement.attr.marker_list, bn);
        }
        break;
    case 0x0d : /* Text Representation (IX, IX, E, R, R, CO)*/
        {
            int idx = r_get_IX(r);
            TextBundle *bn = (TextBundle*)cgm_get_bundle(&c->pElement.attr.text_list, idx);
            if(!bn){
                bn = loa_array_append(&c->pElement.attr.text_list, 1);
                bn->base.idx = idx;
            }
            bn->nTextFontIndex = r_get_IX(r);
            bn->eTextPrecision = r_get_text_precision(r);
            bn->nCharacterSpacing = r_get_R(r);
            bn->nCharacterExpansion = r_get_R(r);
            bn->base.color = r_get_color(r, &c->pElement, 0);
            c->pElement.attr.text_bundled = loa_array_index(&c->pElement.attr.text_list, bn);
        }
        break;
    case 0x0e : /* Fill Representation (IX, E, CO, IX, IX) */
        {
            int idx = r_get_IX(r);
            FillBundle *bn = (FillBundle*)cgm_get_bundle(&c->pElement.attr.fill_list, idx);
            if(!bn){
                bn = loa_array_append(&c->pElement.attr.fill_list, 1);
                bn->base.idx = idx;
            }
            bn->eFillInteriorStyle = r_get_interior_style(r);
            bn->base.color = r_get_color(r, &c->pElement, 0);
            bn->nFillPatternIndex = r_get_IX(r);
            bn->nFillHatchIndex = r_get_IX(r);
            c->pElement.attr.fill_bundled = loa_array_index(&c->pElement.attr.fill_list, bn);
        }
        break;
    case 0x0f : /* Edge Representation (IX, IX, SS, CO) */
        {
            int idx = r_get_IX(r);
            LineBundle *bn = (LineBundle*)cgm_get_bundle(&c->pElement.attr.edge_list, idx);
            if(!bn){
                bn = loa_array_append(&c->pElement.attr.edge_list, 1);
                bn->base.idx = idx;
            }
            bn->type = (line_type_t)r_get_IX(r);
            bn->width = r_get_V(r, c->pElement.attr.eEdgeWidthSpecMode);
            bn->base.color = r_get_color(r, &c->pElement, 0);
            c->pElement.attr.edge_bundled = loa_array_index(&c->pElement.attr.edge_list, bn);
        }
        break;
    case 0x10 : /* Interior Style Specification Mode (E) */
        break;    // NS
    case 0x11 : /* Line and Edge Type Definition (IX, SS, nI) */
        break;
    case 0x12 : /* Hatch Style Definition (IX, E, 4SS, SS, I, nI, nIX) */
        break;               // NS
    case 0x13 : /* Geometric Pattern Definition (IX, N, 2P) */
        break;         // NS
    case 0xff : /* inquire VDC EXTENT */
        break;
    case 0xfe : /* inquire Background Color */
        break;
    case 0xfd : /* inquire Device Viewport */
        break;
    case 0xfc : /* set Font Selection Mode */
        break;
    case 0xfb : /* inquire Color Selection Mode */
        break;
    case 0xfa : /* inquire Font Selection Mode */
        break;
    case 0xf9 : /* set Char Height Spec Mode*/
        //ImplGetUI16(); // -Wall is this really needed?
        break;
    case 0xf8 : /* set Background Style */
        break;
    default:
        break;
    }
    return 0;
};
