/*
 * Copyright (C) 2000 – 2020  LibreOffice contributors
 * Copyright (C) 2021         Magnus Bergman <magnus@snisurset.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * This file incorporates work covered by the following license notice:
 *
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements. See the NOTICE file distributed
 *   with this work for additional information regarding copyright
 *   ownership. The ASF licenses this file to you under the Apache
 *   License, Version 2.0 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.apache.org/licenses/LICENSE-2.0 .
 */
#include <debug.h>
#include <stdlib.h>
#include <string.h>

#include "elements.h"


char *strcasestr(const char *haystack, const char *needle);


void font_entry_done(FontEntry *e)
{
    if(e->font_name)
        free(e->font_name);
    if(e->char_set_value)
        free(e->char_set_value);
}


CGMFList *cgm_font_list_init(CGMFList *fl)
{
    fl->nFontNameCount = 0;
    fl->nCharSetCount = 0;
    loa_array_init(&fl->font_entry_list, sizeof(FontEntry));
    return fl;
}


void cgm_font_list_done(CGMFList *fl)
{
    int i;
    for(i=0;i<loa_array_length(&fl->font_entry_list);++i)
        font_entry_done((FontEntry*)loa_array_get(&fl->font_entry_list, i));
    loa_array_done(&fl->font_entry_list);
}


FontEntry* CGMFList_GetFontEntry(CGMFList *fl, uint32_t nIndex )
{
    uint32_t nInd = nIndex;
    if ( nInd )
        nInd--;
    return nInd < loa_array_length(&fl->font_entry_list) ? (FontEntry*)loa_array_get(&fl->font_entry_list, nInd) : NULL;
}


void CGMFList_InsertName(CGMFList *fl, const char *pSource, uint32_t nSize )
{
    FontEntry* pFontEntry;
    if (fl->nFontNameCount == loa_array_length(&fl->font_entry_list))
    {
        pFontEntry = (FontEntry*)loa_array_append(&fl->font_entry_list, 1);
        font_entry_init(pFontEntry);
    }
    else
    {
        //pFontEntry = &aFontEntryList[nFontNameCount];
        pFontEntry = (FontEntry*)loa_array_get(&fl->font_entry_list, fl->nFontNameCount);
    }
    fl->nFontNameCount++;

    if (nSize == 0)
        return;

    //std::vector<int8_t> aBuf(pSource, pSource + nSize);
    //int8_t* pFound = ImplSearchEntry(aBuf.data(), reinterpret_cast<int8_t const *>("ITALIC"), nSize, 6);
    char *s = strndup(pSource, nSize);
    char *pFound = strcasestr(s, "ITALIC");
    if (pFound)
    {
        pFontEntry->type |= 1;
        //uint32_t nPrev = pFound - aBuf.data();
        uint32_t nPrev = pFound - s;
        uint32_t nToCopyOfs = 6;
        if ( nPrev && ( pFound[ -1 ] == '-' || pFound[ -1 ] == ' ' ) )
        {
            nPrev--;
            pFound--;
            nToCopyOfs++;
        }
        uint32_t nToCopy = nSize - nToCopyOfs - nPrev;
        if ( nToCopy )
        {
            memmove( pFound, pFound + nToCopyOfs, nToCopy );
        }
        nSize -= nToCopyOfs;
        s[nSize] = '\0';
    }
    //pFound = ImplSearchEntry(aBuf.data(), reinterpret_cast<int8_t const *>("BOLD"), nSize, 4);
    pFound = strcasestr(s, "BOLD");
    if ( pFound )
    {
        pFontEntry->type |= 2;

        uint32_t nPrev = pFound - s;
        uint32_t nToCopyOfs = 4;
        if ( nPrev && ( pFound[ -1 ] == '-' || pFound[ -1 ] == ' ' ) )
        {
            nPrev--;
            pFound--;
            nToCopyOfs++;
        }
        uint32_t nToCopy = nSize - nToCopyOfs - nPrev;
        if ( nToCopy )
        {
            memmove( pFound, pFound + nToCopyOfs, nToCopy );
        }
        nSize -= nToCopyOfs;
        s[nSize] = '\0';
    }
#if 1
    pFontEntry->font_name = s;
#else
    pFontEntry->aFontName.assign(aBuf.data(), aBuf.data() + nSize);
#endif
}


void CGMFList_InsertCharSet(CGMFList *fl, const char *pSource, uint32_t nSize )
{
    FontEntry* pFontEntry;
    if (fl->nCharSetCount == loa_array_length(&fl->font_entry_list))
    {
        pFontEntry = (FontEntry*)loa_array_append(&fl->font_entry_list, 1);
        font_entry_init(pFontEntry);
    }
    else
    {
        pFontEntry = (FontEntry*)loa_array_get(&fl->font_entry_list, fl->nCharSetCount);
    }
    fl->nCharSetCount++;

    if (nSize == 0)
        return;

    pFontEntry->char_set_value = strndup(pSource, nSize);
}



cgm_attributes_t *cgm_attributes_init(cgm_attributes_t *a)
{
    a->nAspectSourceFlags = 0; /* all flags are individual */

    /* Line attributes */
    //line_bundle_init(&a->aLineBundle);
    loa_array_init(&a->line_list, sizeof(LineBundle));
    a->eLineWidthSpecMode = SM_SCALED;
    a->eLineCapType = LCT_NONE;
    a->eLineJoinType = LJT_NONE;
    {
        LineBundle *bn = loa_array_append(&a->line_list, 1);
        bn->base.idx = 1;
        bn->base.color = 0xffffff;
        bn->type = LT_SOLID;
        bn->width = 1;
    }
    a->line_bundled = 0;

    /* Marker attributes */
    loa_array_init(&a->marker_list, sizeof(MarkerBundle));
    a->eMarkerSizeSpecMode = SM_SCALED;
    {
        MarkerBundle *bn = loa_array_append(&a->marker_list, 1);
        bn->base.idx = 1;
        bn->base.color = 0xffffff;
        bn->eMarkerType = MT_STAR; /* or MT_POINT ?? */
        bn->nMarkerSize = 1;
    }
    a->marker_bundled = 0;

    /* Edge attributes */
    loa_array_init(&a->edge_list, sizeof(LineBundle));
    a->eEdgeVisibility = EV_OFF;
    a->eEdgeWidthSpecMode = SM_SCALED;
    {
        LineBundle *bn = loa_array_append(&a->edge_list, 1);
        bn->base.idx = 1;
        bn->base.color = 0xffffff;
        bn->type = LT_SOLID; /* or LT_NONE ?? */
        bn->width = 1;
    }
    a->edge_bundled = 0;

    /* Text attributes */
    loa_array_init(&a->text_list, sizeof(TextBundle));
    a->nCharacterHeight = 327;
    a->nCharacterOrientation[0] = 0;
    a->nCharacterOrientation[1] = 1;
    a->nCharacterOrientation[2] = 1;
    a->nCharacterOrientation[3] = 0;
    a->eUnderlineMode = UM_OFF;
    a->nUnderlineColor = 0xffffff;
    a->eTextPath = TPR_RIGHT;
    a->eTextAlignmentH = TAH_NORMAL;
    a->eTextAlignmentV = TAV_NORMAL;
    a->nCharacterSetIndex = a->nAlternateCharacterSetIndex = 1;
    a->eCharacterCodingA = CCA_BASIC_7;
    {
        TextBundle *bn = loa_array_append(&a->text_list, 1);
        bn->base.idx = 1;
        bn->base.color = 0xffffff;
        bn->nTextFontIndex = 1;
        bn->eTextPrecision = TPR_STRING; /* TPR_UNDEFINED ?? */
        bn->nCharacterExpansion = 1;
        bn->nCharacterSpacing = 0;
    }
    a->text_bundled = 0;
    cgm_font_list_init(&a->aFontList);
    a->nTextAlignmentVCont = a->nTextAlignmentHCont = 0.0;

    /* Fill attributes */
    loa_array_init(&a->fill_list, sizeof(FillBundle));
    {
        FillBundle *bn = loa_array_append(&a->fill_list, 1);
        bn->base.idx = 1;
        bn->base.color = 0xffffff;
        bn->eFillInteriorStyle = FIS_HOLLOW;
        bn->nFillHatchIndex = 1;
        bn->nFillPatternIndex = 1;
    }
    a->fill_bundled = 0;

    return a;
}


void cgm_attributes_done(cgm_attributes_t *a)
{
    loa_array_done(&a->line_list);
    loa_array_done(&a->marker_list);
    loa_array_done(&a->edge_list);
    loa_array_done(&a->text_list);
    cgm_font_list_done(&a->aFontList);
    loa_array_done(&a->fill_list);
}


CGMElements *cgm_element_init(CGMElements *e)
{
#if 0
    nIntegerPrecision = nIndexPrecision = 2;
    nRealSize = nVDCRealSize = 4;
    nColorIndexPrecision = 1;
    nColorPrecision = 1;
    nVDCIntegerPrecision = 2;
    eRealPrecision = eVDCRealPrecision = RP_FIXED;      //RP_FLOAT;
#endif

    //e->nMetaFileVersion = 1;
    e->eScalingMode = SM_ABSTRACT;
#if 0
    e->eVDCType = VDC_INTEGER;
#endif
#if 1
    e->aVDCExtent.pos.x = e->aVDCExtent.pos.y = 0;
    e->aVDCExtent.size.x = e->aVDCExtent.size.y = 1.0;
    e->aVDCExtentMaximum.pos.x = e->aVDCExtentMaximum.pos.y = 0;
    e->aVDCExtentMaximum.size.x = e->aVDCExtentMaximum.size.y = 1.0;
#else
    e->aVDCExtent.Left = e->aVDCExtent.Bottom = 0;
//  e->aVDCExtent.Right = e->aVDCExtent.Top = 32767;
    e->aVDCExtent.Right = e->aVDCExtent.Top = 1.0;
    e->aVDCExtentMaximum.Left = e->aVDCExtentMaximum.Bottom = 0;
//  e->aVDCExtentMaximum.Right = e->aVDCExtentMaximum.Top = 32767;
    e->aVDCExtentMaximum.Right = e->aVDCExtentMaximum.Top = 1.0;
#endif

    //e->eDeviceViewPortMode = DVPM_FRACTION;
    e->nDeviceViewPortScale = 1;
    e->eDeviceViewPortMap = DVPM_FORCED;
    e->eDeviceViewPortMapH = DVPMH_LEFT;
    e->eDeviceViewPortMapV = DVPMV_BOTTOM;
#if 1
    e->aDeviceViewPort.pos.x = 0;
    e->aDeviceViewPort.pos.y = 1;
    e->aDeviceViewPort.size.x = 0;
    e->aDeviceViewPort.size.y = 0;
#else
    e->aDeviceViewPort.Left = 0;
    e->aDeviceViewPort.Top = 1;
    e->aDeviceViewPort.Right = 0;
    e->aDeviceViewPort.Bottom = 1;
#endif

    e->nMitreLimit = 32767;
    e->eClipIndicator = CI_ON;
    e->aClipRect = e->aVDCExtent;

    e->eColorSelectionMode = CSM_INDEXED;
    e->nColorMaximumIndex = 63;
    int i;
    for ( i = 0; i < 256; e->aColorTableEntryIs[ i++ ] = 0 ) ;
    e->aColorTable[ 0 ] = 0;
    for ( i = 1; i < 256; e->aColorTable[ i++ ] = 0xffffff ) ;
    e->nLatestColorMaximumIndex = 63;
    e->aLatestColorTable[ 0 ] = 0;
    for ( i = 1; i < 256; e->aLatestColorTable[ i++ ] = 0xffffff ) ;
    e->nColorValueExtent[ 0 ] = e->nColorValueExtent[ 1 ] = e->nColorValueExtent[ 2 ] = 0;
    e->nColorValueExtent[ 3 ] = e->nColorValueExtent[ 4 ] = e->nColorValueExtent[ 5 ] = 255;

    CGMElements_ImplInsertHatch(e, 0, 0, 0, 0 );
    CGMElements_ImplInsertHatch(e, 1, 0, 125, 0 );
    CGMElements_ImplInsertHatch(e, 2, 0, 125, 900 );
    CGMElements_ImplInsertHatch(e, 3, 0, 125, 450 );
    CGMElements_ImplInsertHatch(e, 4, 0, 125, 1350 );
    CGMElements_ImplInsertHatch(e, 5, 1, 125, 0 );
    CGMElements_ImplInsertHatch(e, 6, 1, 125, 450 );
    CGMElements_ImplInsertHatch(e, -1, 0, 75, 0 );
    CGMElements_ImplInsertHatch(e, -2, 0, 75, 900 );
    CGMElements_ImplInsertHatch(e, -3, 0, 75, 450 );
    CGMElements_ImplInsertHatch(e, -4, 0, 75, 1350 );
    CGMElements_ImplInsertHatch(e, -5, 1, 75, 0 );
    CGMElements_ImplInsertHatch(e, -6, 1, 75, 450 );
    CGMElements_ImplInsertHatch(e, -7, 2, 125, 0 );
    CGMElements_ImplInsertHatch(e, -8, 2, 125, 900 );
    CGMElements_ImplInsertHatch(e, -9, 2, 125, 450 );
    CGMElements_ImplInsertHatch(e, -10, 2, 125, 1350 );
    CGMElements_ImplInsertHatch(e, -11, 0, 40, 0 );
    CGMElements_ImplInsertHatch(e, -12, 0, 40, 900 );
    CGMElements_ImplInsertHatch(e, -13, 0, 40, 450 );
    CGMElements_ImplInsertHatch(e, -14, 0, 40, 1350 );
    CGMElements_ImplInsertHatch(e, -15, 1, 40, 0 );
    CGMElements_ImplInsertHatch(e, -16, 1, 40, 900 );
    CGMElements_ImplInsertHatch(e, -21, 0, 250, 0 );
    CGMElements_ImplInsertHatch(e, -22, 0, 250, 900 );
    CGMElements_ImplInsertHatch(e, -23, 0, 250, 450 );
    CGMElements_ImplInsertHatch(e, -24, 0, 250, 1350 );
    CGMElements_ImplInsertHatch(e, -25, 1, 250, 0 );
    CGMElements_ImplInsertHatch(e, -26, 1, 250, 450 );

    cgm_attributes_init(&e->attr);

    e->eTransparency = T_ON;

    e->nBackGroundColor = e->nAuxiliaryColor = 0;

    e->bSegmentCount = false;

    e->nScalingFactor = 1.0;

    return e;
}


void cgm_element_done(CGMElements *e)
{
    cgm_attributes_done(&e->attr);
}


#if 0
CGMElements& CGMElements::operator=( const CGMElements& rSource )
{
    if (this == &rSource)
        return *this;

    uint32_t nIndex;

#if 0
    nVDCIntegerPrecision = rSource.nVDCIntegerPrecision;
    nIntegerPrecision = rSource.nIntegerPrecision;
    eRealPrecision = rSource.eRealPrecision;
    nRealSize = rSource.nRealSize;
    nIndexPrecision = rSource.nIndexPrecision;
    nColorPrecision = rSource.nColorPrecision;
    nColorIndexPrecision = rSource.nColorIndexPrecision;
#endif

    nMetaFileVersion = rSource.nMetaFileVersion;
    eScalingMode = rSource.eScalingMode;
    nScalingFactor = rSource.nScalingFactor;
#if 0
    eVDCType = rSource.eVDCType;
    eVDCRealPrecision = rSource.eVDCRealPrecision;
    nVDCRealSize = rSource.nVDCRealSize;
#endif
    aVDCExtent = rSource.aVDCExtent;
    aVDCExtentMaximum = rSource.aVDCExtentMaximum;
    eDeviceViewPortMode = rSource.eDeviceViewPortMode;
    nDeviceViewPortScale = rSource.nDeviceViewPortScale;
    eDeviceViewPortMap = rSource.eDeviceViewPortMap;
    eDeviceViewPortMapH = rSource.eDeviceViewPortMapH;
    eDeviceViewPortMapV = rSource.eDeviceViewPortMapV;
    aDeviceViewPort = rSource.aDeviceViewPort;
    nMitreLimit = rSource.nMitreLimit;
    eClipIndicator = rSource.eClipIndicator;
    aClipRect = rSource.aClipRect;
    eColorSelectionMode = rSource.eColorSelectionMode;
    nColorMaximumIndex = rSource.nColorMaximumIndex;
    nLatestColorMaximumIndex = rSource.nLatestColorMaximumIndex;

    for ( nIndex = 1; nIndex < 256; nIndex++ )      // do not overwrite the background color
    {
        aColorTableEntryIs[ nIndex ] = rSource.aColorTableEntryIs[ nIndex ];
        aColorTable[ nIndex ] = rSource.aColorTable[ nIndex ];
        aLatestColorTable[ nIndex ] = rSource.aColorTable[ nIndex ];
    }

    for ( nIndex = 0; nIndex < 8; nIndex++ )
    {
        nColorValueExtent[ nIndex ] = rSource.nColorValueExtent[ nIndex ];
    }
    nAspectSourceFlags = rSource.nAspectSourceFlags;

    CopyAllBundles( rSource.aLineList, aLineList );
    aLineBundle = rSource.aLineBundle;
    pLineBundle = static_cast<LineBundle*>(GetBundle( aLineList, rSource.pLineBundle->GetIndex() ));
    eLineWidthSpecMode = rSource.eLineWidthSpecMode;
    eLineCapType = rSource.eLineCapType;
    eLineJoinType = rSource.eLineJoinType;

    CopyAllBundles( rSource.aMarkerList, aMarkerList );
    aMarkerBundle = rSource.aMarkerBundle;
    pMarkerBundle = static_cast<MarkerBundle*>(GetBundle( aMarkerList, rSource.pMarkerBundle->GetIndex() ));
    eMarkerSizeSpecMode = rSource.eMarkerSizeSpecMode;

    CopyAllBundles( rSource.aEdgeList, aEdgeList );
    aEdgeBundle = rSource.aEdgeBundle;
    pEdgeBundle = static_cast<EdgeBundle*>(GetBundle( aEdgeList, rSource.pEdgeBundle->GetIndex() ));
    eEdgeVisibility = rSource.eEdgeVisibility;
    eEdgeWidthSpecMode = rSource.eEdgeWidthSpecMode;

    CopyAllBundles( rSource.aTextList, aTextList );
    aTextBundle = rSource.aTextBundle;
    pTextBundle = static_cast<TextBundle*>(GetBundle( aTextList, rSource.pTextBundle->GetIndex() ));
    nCharacterHeight = rSource.nCharacterHeight;
    nCharacterOrientation[ 0 ] = rSource.nCharacterOrientation[ 0 ];
    nCharacterOrientation[ 1 ] = rSource.nCharacterOrientation[ 1 ];
    nCharacterOrientation[ 2 ] = rSource.nCharacterOrientation[ 2 ];
    nCharacterOrientation[ 3 ] = rSource.nCharacterOrientation[ 3 ];
    eUnderlineMode = rSource.eUnderlineMode;
    nUnderlineColor = rSource.nUnderlineColor;
    eTextPath = rSource.eTextPath;
    eTextAlignmentH = rSource.eTextAlignmentH;
    eTextAlignmentV = rSource.eTextAlignmentV;
    nTextAlignmentHCont = rSource.nTextAlignmentHCont;
    nTextAlignmentVCont = rSource.nTextAlignmentVCont;
    nCharacterSetIndex = rSource.nCharacterSetIndex;
    nAlternateCharacterSetIndex = rSource.nAlternateCharacterSetIndex;
    aFontList = rSource.aFontList;
    eCharacterCodingA = rSource.eCharacterCodingA;

    CopyAllBundles( rSource.aFillList, aFillList );
    aFillBundle = rSource.aFillBundle;
    pFillBundle = static_cast<FillBundle*>(GetBundle( aFillList, rSource.pFillBundle->GetIndex() ));
    aFillRefPoint = rSource.aFillRefPoint;
    eTransparency = rSource.eTransparency;
    nAuxiliaryColor = rSource.nAuxiliaryColor;

    maHatchMap = rSource.maHatchMap;
    bSegmentCount = rSource.bSegmentCount;
    return *this;
}
#endif


void CGMElements_ImplInsertHatch(CGMElements *e, int32_t nKey, int nStyle, long nDistance, long nAngle )
{
#if 1
    DEBUG_p("not implemented");
    /* FIXME!! */
#else
    HatchEntry& rEntry = maHatchMap[nKey];
    rEntry.HatchStyle = nStyle;
    rEntry.HatchDistance = nDistance;
    rEntry.HatchAngle = nAngle;
#endif
}


#if 0
void CGMElements_CopyAllBundles(const array_t *rSource, array_t *rDest )
{
#if 1
    DEBUG_p("not implemented");
    /* FIXME!! */
#else
    rDest.clear();

    for (auto & pPtr : rSource)
    {
        rDest.push_back( pPtr->Clone() );
    }
#endif
};
#endif


Bundle* cgm_get_bundle(loa_array_t *rList, int idx)
{
    DEBUG_p("index: %d",(int)idx);
    int i;
    for(i=0;i<loa_array_length(rList);++i){
        Bundle *bn = (Bundle*)loa_array_get(rList, i);
        if(bn->idx == idx)
            return bn;
    }
    return NULL;
}
