/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>
#include <string.h>

#include "abydos-archive.h"


#define START_PAGES 512
#define BUFFER_SIZE 4096


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    abydos_archive_type_t type;
    abydos_t **page;
};


static const char **
_comics_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef USE_ACEUNPACK
        "application/x-cba",
#endif
        "application/vnd.comicbook-rar", "application/x-cbr", "application/x-rar",
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
        "application/vnd.comicbook+zip", "application/x-cbz", "application/zip",
        "application/x-cb7", "application/x-7z-compressed",
        "application/x-cbt", "application/x-tar",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_comics_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    if(!strcasecmp(mime_type, "application/x-cb7") ||
       !strcasecmp(mime_type, "application/x-7z-compressed"))
    {
        h->type = AT_7Z;
    }else
#endif
#ifdef USE_ACEUNPACK
    if(!strcasecmp(mime_type, "application/x-cba")){
        h->type = AT_ACE;
    }else
#endif
    if(!strcasecmp(mime_type, "application/vnd.comicbook-rar") ||
       !strcasecmp(mime_type, "application/x-cbr") ||
       !strcasecmp(mime_type, "application/x-rar"))
    {
        h->type = AT_RAR;
    }else
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    if(!strcasecmp(mime_type, "application/x-cbt") ||
       !strcasecmp(mime_type, "application/x-tar"))
    {
        h->type = AT_TAR;
    }else
    if(!strcasecmp(mime_type, "application/vnd.comicbook+zip") ||
       !strcasecmp(mime_type, "application/x-cbz") ||
       !strcasecmp(mime_type, "application/zip"))
    {
        h->type = AT_ZIP;
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_comics_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            abydos_destroy(h->page[i]);
        free(h->page);
    }
    free(h);
}


static int
_comics_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    abydos_archive_t archive;
    int result = -1;
    if(!abydos_archive_init(&archive, h->type, filename))
        return -1;

    h->info->page_count = 0;
    h->page = malloc(START_PAGES * sizeof(abydos_t*));
    for(;;){
        char buffer[BUFFER_SIZE];
        int bytes;
        abydos_t *ar;
        if(abydos_archive_next(&archive) < 0)
            break;
        bytes = abydos_archive_entry_read(&archive, buffer, BUFFER_SIZE);
        if(bytes < 4)
            continue;
        if(h->info->page_count + 1 > START_PAGES)
            h->page = realloc(h->page, (h->info->page_count + 1) * sizeof(abydos_t*));
        /* try to detect the image type */
        if(!strncmp(buffer, "\377\330\377", 3)){
            ar = h->page[h->info->page_count] = abydos_create("image/jpeg");
            if(!ar){
                /* JPEG not supported */
                return -1;
            }
        }else{
            continue;
        }
        abydos_load_begin(ar, NULL, NULL, NULL);
        for(;;){
            if(abydos_load_feed(ar, buffer, bytes) < 0)
                goto decode_error;
            bytes = abydos_archive_entry_read(&archive, buffer, BUFFER_SIZE);
            DEBUG_p("got %d bytes", (int)bytes);
            if(bytes <= 0)
                break;
        }
        if(abydos_load_end(ar) < 0){
decode_error:
            abydos_destroy(ar);
            continue;
        }
        ++h->info->page_count;
    }
    if(h->info->page_count > 0){
        if(h->info->page_count < START_PAGES)
            h->page = realloc(h->page, h->info->page_count * sizeof(abydos_t*));
        result = 0;
    }else{
        free(h->page);
        h->page = NULL;
    }
    abydos_archive_done(&archive);
    return result;
}


static cairo_surface_t *
_comics_get_image_surface(abydos_plugin_handle_t *h,
                          int page,
                          cairo_bool_t *layer_visibility,
                          int variant,
                          double time,
                          int frame)
{
    return abydos_get_image_surface(h->page[page]);
}


static void
_comics_render(abydos_plugin_handle_t *h,
               cairo_t *cr,
               int page,
               cairo_bool_t *layer_visibility,
               int variant,
               double time,
               int frame)
{
    abydos_render(h->page[page], cr);
}


static void
_comics_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *p_info)
{
    abydos_info_t a_info;
    abydos_get_info(h->page[page], &a_info);
    p_info->width = a_info.width;
    p_info->height = a_info.height;
    p_info->pixel_ratio = a_info.pixel_ratio;
    p_info->layer_count = abydos_get_layer_count(h->page[page]);
    p_info->variant_count = abydos_get_variant_count(h->page[page]);
    p_info->frame_count = abydos_get_frame_count(h->page[page]);
    //p_info->default_frame = abydos_get_default_frame(h->page[page]);
    p_info->scalable_size = a_info.features & ABYDOS_FEATURE_SCALABLE_SIZE;
    p_info->scalable_time = a_info.features & ABYDOS_FEATURE_SCALABLE_TIME;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _comics_get_mime_types,
        .new = _comics_new,
        .free = _comics_free,
        .create_from_file = _comics_create_from_file,
        .get_image_surface = _comics_get_image_surface,
        .render = _comics_render,
        .get_page_info = _comics_get_page_info
    };
    return &p;
}
