/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <string.h>

#include "debug.h"
#include "nil-fax.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#define ERROR -1
#define OUT_OF_DATA -2
#define END_OF_LINE -3

typedef struct{
    const uint8_t *data;
    const uint8_t *end;
    int bits;
}bitstream_t;


static void
_bitstream_init(bitstream_t *bs, const uint8_t *data, size_t len)
{
    bs->data = data;
    bs->end = data+len;
    bs->bits = 0;
}


static int
_bitstream_get_bit(bitstream_t *bs)
{
    int bit;
    if(!(bs->bits&0xff)){
        if(bs->data >= bs->end){
            DEBUG_p("out of data");
            return OUT_OF_DATA;
        }
        bs->bits = (*bs->data++)<<1 | 1;
    }
    bit = bs->bits&256;
    bs->bits<<=1;
    return bit;
}


typedef struct{
    bitstream_t bitstream;
    const uint8_t *tree;
    int terminate;
}huffman_t;


static void
_huffman_init(huffman_t *huff, const uint8_t *data, size_t len)
{
    _bitstream_init(&huff->bitstream, data, len);
    huff->tree = NULL;
    huff->terminate = 0;
}


static int
_huffman_sync(huffman_t *huff)
{
    huff->terminate = 0;
    for(;;){
        int bit = _bitstream_get_bit(&huff->bitstream);
        if(bit < 0)
            return bit;
        if(bit)
            return 0;
    }
}


static int
_huffman_get(huffman_t *huff)
{
    const uint8_t *node;
    int val;
    node = huff->tree;
    for(;;){
        int bit;
        val = *node;
        if(127 == val){
            DEBUG_p("end of line");
            return END_OF_LINE;
        }
        if(val < 128)
            break;
        bit = _bitstream_get_bit(&huff->bitstream);
        if(bit < 0)
            return bit;
        ++node;
        if(!bit)
            node += val & ~128;
    }
    huff->terminate = val < 64;
    if(val > 64)
        val = 64 + (val-64) * 64;
    return val;
}


typedef struct{
    huffman_t huffman;
    int black;
}rle_t;


static void
_rle_init(rle_t *rle, const uint8_t *data, size_t len)
{
    _huffman_init(&rle->huffman, data, len);
    rle->black = 0;
}


static void
_rle_sync(rle_t *rle)
{
    _huffman_sync(&rle->huffman);
    rle->black = 0;
}


static int
_rle_get(rle_t *rle)
{
    static const uint8_t white_tree[] = {
        0x99, 0x8b, 0x83, 0x81, 0x07, 0x06, 0x85, 0x81,
        0x40, 0x81, 0x0f, 0x0e, 0x05, 0x87, 0x81, 0x04,
        0x83, 0x81, 0x11, 0x10, 0x09, 0x83, 0x81, 0x08,
        0x41, 0x03, 0xcd, 0xa5, 0x81, 0x02, 0x97, 0x89,
        0x81, 0x43, 0x83, 0x81, 0x55, 0x54, 0x81, 0x53,
        0x52, 0x87, 0x83, 0x81, 0x51, 0x50, 0x81, 0x4f,
        0x4e, 0x83, 0x81, 0x4d, 0x4c, 0x48, 0x89, 0x85,
        0x81, 0x49, 0x81, 0x4b, 0x4a, 0x81, 0x47, 0x46,
        0x59, 0x95, 0x89, 0x81, 0x42, 0x83, 0x81, 0x3a,
        0x39, 0x81, 0x38, 0x37, 0x85, 0x81, 0x19, 0x81,
        0x34, 0x33, 0x83, 0x81, 0x32, 0x31, 0x18, 0x8f,
        0x89, 0x81, 0x12, 0x83, 0x81, 0x5a, 0x58, 0x81,
        0x57, 0x56, 0x83, 0x81, 0x3c, 0x3b, 0x1b, 0x0b,
        0xa5, 0x8f, 0x81, 0x0a, 0x87, 0x83, 0x81, 0x45,
        0x44, 0x81, 0x00, 0x3f, 0x83, 0x81, 0x3e, 0x3d,
        0x1c, 0x8d, 0x85, 0x81, 0x15, 0x81, 0x2c, 0x2b,
        0x83, 0x81, 0x2a, 0x29, 0x81, 0x28, 0x27, 0x85,
        0x81, 0x1a, 0x81, 0x36, 0x35, 0x0c, 0x95, 0x87,
        0x81, 0x01, 0x83, 0x81, 0x20, 0x1f, 0x13, 0x87,
        0x83, 0x81, 0x26, 0x25, 0x81, 0x24, 0x23, 0x83,
        0x81, 0x22, 0x21, 0x14, 0x87, 0x81, 0x0d, 0x83,
        0x81, 0x30, 0x2f, 0x17, 0x85, 0x81, 0x16, 0x81,
        0x2e, 0x2d, 0x83, 0x81, 0x1e, 0x1d, 0x99, 0x8b,
        0x87, 0x83, 0x81, 0x67, 0x66, 0x81, 0x65, 0x64,
        0x81, 0x5d, 0x5c, 0x87, 0x83, 0x81, 0x63, 0x62,
        0x81, 0x61, 0x60, 0x83, 0x81, 0x5f, 0x5e, 0x5b,
        0x7f,
    };
    static const uint8_t black_tree[] = {
        0x83, 0x81, 0x02, 0x03, 0x83, 0x81, 0x04, 0x01,
        0x83, 0x81, 0x05, 0x06, 0x85, 0x81, 0x07, 0x81,
        0x08, 0x09, 0xab, 0xa7, 0x81, 0x0c, 0x95, 0x87,
        0x81, 0x00, 0x83, 0x81, 0x2b, 0x2a, 0x15, 0x87,
        0x83, 0x81, 0x27, 0x26, 0x81, 0x25, 0x24, 0x83,
        0x81, 0x23, 0x22, 0x14, 0x8d, 0x85, 0x81, 0x13,
        0x81, 0x1d, 0x1c, 0x83, 0x81, 0x1b, 0x1a, 0x81,
        0x42, 0x41, 0x0f, 0x81, 0x0b, 0x0a, 0xb3, 0x99,
        0x81, 0x0e, 0x8d, 0x85, 0x81, 0x16, 0x81, 0x29,
        0x28, 0x83, 0x81, 0x21, 0x20, 0x81, 0x1f, 0x1e,
        0x87, 0x83, 0x81, 0x3f, 0x3e, 0x81, 0x31, 0x30,
        0x11, 0x97, 0x89, 0x81, 0x10, 0x83, 0x81, 0x43,
        0x3d, 0x81, 0x3a, 0x39, 0x87, 0x83, 0x81, 0x2f,
        0x2e, 0x81, 0x2d, 0x2c, 0x83, 0x81, 0x33, 0x32,
        0x17, 0x0d, 0xc3, 0xa1, 0x8f, 0x81, 0x40, 0x87,
        0x83, 0x81, 0x52, 0x51, 0x81, 0x50, 0x4f, 0x83,
        0x81, 0x4e, 0x4d, 0x36, 0x89, 0x85, 0x81, 0x35,
        0x81, 0x48, 0x47, 0x81, 0x46, 0x45, 0x85, 0x81,
        0x44, 0x81, 0x5a, 0x59, 0x19, 0x93, 0x87, 0x81,
        0x18, 0x83, 0x81, 0x58, 0x57, 0x3c, 0x85, 0x81,
        0x3b, 0x81, 0x56, 0x55, 0x83, 0x81, 0x54, 0x53,
        0x38, 0x8b, 0x85, 0x81, 0x37, 0x81, 0x4c, 0x4b,
        0x83, 0x81, 0x4a, 0x49, 0x34, 0x12, 0x99, 0x8b,
        0x87, 0x83, 0x81, 0x67, 0x66, 0x81, 0x65, 0x64,
        0x81, 0x5d, 0x5c, 0x87, 0x83, 0x81, 0x63, 0x62,
        0x81, 0x61, 0x60, 0x83, 0x81, 0x5f, 0x5e, 0x5b,
        0x7f,
    };
    for(;;){
        int c;
        if(rle->huffman.terminate)
            rle->black = !rle->black;
        if(rle->black){
            rle->huffman.tree = black_tree;
        }else{
            rle->huffman.tree = white_tree;
        }
        c = _huffman_get(&rle->huffman);
        if(0 == c)
            continue;
        return c;
    }
}


int nil_fax_decode(nile_t *ni, nil_fax_compression_t compression, const uint8_t *data, size_t len)
{
    rle_t rle;
    uint8_t *dst = ni->pixels;
    int skip = nile_get_stride(ni) - ni->width;
    int y;
    if(compression != NIL_FAX_MH)
        return -1;
    _rle_init(&rle, data, len);
    //ni->height = 114;
    _rle_sync(&rle);
    for(y=0;y<ni->height;++y){
        int left = ni->width;
        DEBUG_p("line %d", y);
        for(;;){
            int c = _rle_get(&rle);
            if(c < 0){
                if(OUT_OF_DATA == c){
                    ni->height = y;
                    if(left < ni->width){
                        memset(dst, 1, left);
                        ++ni->height;
                    }
                    return 0;
                }else if(END_OF_LINE == c){
                    DEBUG_p("eol");
                    _rle_sync(&rle);
                    if(left > 0 && left < ni->width){
                        memset(dst, 1, left);
                        ++ni->height;
                        dst += left;
                    }
                    break;
                }else{
                    return -1;
                }
            }
            if(left <= 0)
                continue;
            if(c > left)
                c = left;
            memset(dst, rle.black ? 0 : 1, c);
            dst += c;
            left -= c;
        }
        dst += skip;
    }
    return 0;
}
/*

__eol_______
000000000001

_w0_____ _b1600_______ _b11___ _w1___ _b1 _w1___ _b1 _w1___ _b1 _w3_ _b64______ _b44________ _eol________
00110101 0000001011011 0000101 000111 010 000111 010 000111 010 1000 0000001111 000001010100 000000000001

_w0_____ _b384_______ _b46________ _w1___
00110101 000000110100 000001010110 000111 0001100011100011000111000110001110001100011100011
0001111000011110000111100001111000011110000111100001111000011110

*/
