/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <gd.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */
#include <unistd.h>                     /* unlink() */

/* #define ENABLE_BMP */ /* not working */
#define ENABLE_PNG
#define ENABLE_WBMP
/* #define ENABLE_WEBP */ /* not working */

/*

Supported formats

* BMP
  GD     MIME-type?
  GD2    MIME-type?
  GIF
  JPEG
* PNG
  TGA
  TIFF
* WBMP
* WebP
* XBM
* XPM

*/

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    gdImagePtr(*create_from_data)(int,void*);
    gdImagePtr(*create_from_stream)(FILE*);
    gdImagePtr(*create_from_file)(char*);
    cairo_surface_t *surface;
};


static const char **
_gd_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_BMP
        "image/bmp",
#endif
#ifdef ENABLE_PNG
        "image/png",
#endif
#ifdef ENABLE_WBMP
        "image/vnd.wap.wbmp", "image/x-wireless-bitmap",
#endif
#ifdef ENABLE_WEBP
        "image/webp",
#endif
#ifdef LIBGD_HAS_XBM
        "image/x-xbitmap", "image/x-xbmi",
#endif
#ifdef LIBGD_HAS_XPM
        "image/x-xpixmap", "image/x-xpmi",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gd_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_BMP
    if(!strcasecmp(mime_type,"image/bmp"){
        h->create_from_data = gdImageCreateFromBmpPtr;
        h->create_from_stream = gdImageCreateFromBmp;
        h->create_from_file = NULL;
    }else
#endif
#ifdef ENABLE_PNG
    if(!strcasecmp(mime_type,"image/png")){
        h->create_from_data = gdImageCreateFromPngPtr;
        h->create_from_stream = gdImageCreateFromPng;
        h->create_from_file = NULL;
    }else
#endif
#ifdef ENABLE_WBMP
    if(!strcasecmp(mime_type,"image/vnd.wap.wbmp") ||
       !strcasecmp(mime_type,"image/x-wireless-bitmap"))
    {
        h->create_from_data = gdImageCreateFromWBMPPtr;
        h->create_from_stream = gdImageCreateFromWBMP;
        h->create_from_file = NULL;
    }else
#endif
#ifdef ENABLE_WEBP
    if(!strcasecmp(mime_type,"image/webp")){
        h->create_from_data = gdImageCreateFromWebpPtr;
        h->create_from_stream = gdImageCreateFromWebp;
        h->create_from_file = NULL;
    }else
#endif
#ifdef LIBGD_HAS_XBM
    if(!strcasecmp(mime_type,"image/x-xbitmap")|
       !strcasecmp(mime_type,"image/x-xbmi"))
    {
        h->create_from_data = NULL;
        h->create_from_stream = gdImageCreateFromXbm;
        h->create_from_file = NULL;
    }else
#endif
#ifdef LIBGD_HAS_XPM
    if(!strcasecmp(mime_type,"image/x-xpixmap")|
       !strcasecmp(mime_type,"image/x-xpmi"))
    {
        h->create_from_data = NULL;
        h->create_from_stream = NULL;
        h->create_from_file = gdImageCreateFromXpm;
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_gd_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_surface_t *
_surface_from_image(gdImagePtr im,int *width, int *height)
{
    cairo_surface_t *surface;
    uint32_t *dst;
    int rowskip;
    int x;
    int y;
    *width = gdImageSX(im);
    *height = gdImageSY(im);
    //DEBUG_p("is true color: %s",gdImageTrueColor(im)?"yes":"no");
    //DEBUG_p("has alpha: %s","?");
    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,*width,*height);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-*width;
    for(y=0;y<*height;++y){
        for(x=0;x<*width;++x){
            int pix = gdImageGetTrueColorPixel(im,x,y);
            int alpha = gdAlphaMax - gdTrueColorGetAlpha(pix);
            *dst++ = ((alpha*gdTrueColorGetRed(pix)/gdAlphaMax)<<16)|
                     ((alpha*gdTrueColorGetGreen(pix)/gdAlphaMax)<<8)|
                     ((alpha*gdTrueColorGetBlue(pix)/gdAlphaMax)<<0)|
                     (255*alpha/gdAlphaMax)<<24;
        }
        dst += rowskip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_gd_create_from_stream(abydos_plugin_handle_t *h,FILE *f){
    gdImagePtr im = h->create_from_stream(f);
    if(!im){
        h->info->error = "Failed to creage GD image";
        return -1;
    }
    h->surface = _surface_from_image(im,&h->info->width,&h->info->height);
    gdImageDestroy(im);
    return 0;
}


static int
_open_tmpfile(char **r_filename,FILE** r_f)
{
    static const char template[] = PACKAGE "-XXXXXX";
    const char *tmpdir;
    size_t len;
    char *filename;
    int fd;
    tmpdir = getenv("XDG_RUNTIME_DIR");
    if(!tmpdir)
        tmpdir = getenv("TMPDIR");
    if(!tmpdir)
        tmpdir = "/tmp";
    len = strlen(tmpdir);
    filename = (char*)malloc(len+1+sizeof(template));
    memcpy(filename,tmpdir,len);
    filename[len] = '/';
    memcpy(filename+len+1,template,sizeof(template));
    fd = mkstemp(filename);
    if(fd<0){
        free(filename);
        return 0;
    }
    *r_filename = filename;
    *r_f = fdopen(fd,"w+");
    return 1;
}


static int
_gd_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    if(h->create_from_data){
        gdImagePtr im = h->create_from_data(len,(void*)data);
        if(!im){
            h->info->error = "Failed to creage GD image";
            return -1;
        }
        h->surface = _surface_from_image(im,&h->info->width,&h->info->height);
        return 0;
    }else if(h->create_from_stream){
        FILE *f;
        int res;
        f = fmemopen((void*)data,len,"r");
        if(!f)
            return -1;
        res = _gd_create_from_stream(h,f);
        fclose(f);
        return res;
    }else{
        FILE *f;
        char *tmpfile;
        int res;
        if(_open_tmpfile(&tmpfile,&f)<0)
            return -1;
        res = fwrite(data,1,len,f) < len ? -1 : 0;
        fclose(f);
        if(0 == res){
            gdImagePtr im;
            im = h->create_from_file(tmpfile);
            if(im){
                h->surface = _surface_from_image(im, &h->info->width, &h->info->height);
            }else{
                h->info->error = "Failed to creage GD image";
                res = -1;
            }
        }
        unlink(tmpfile);
        free(tmpfile);
        return res;
    }
}


static int
_gd_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    DEBUG_p("filename: %s",filename);
    if(h->create_from_file){
        gdImagePtr im = h->create_from_file((char*)filename);
        if(!im){
            h->info->error = "Failed to creage GD image";
            return -1;
        }
        h->surface = _surface_from_image(im,&h->info->width,&h->info->height);
        return 0;
    }else{
        FILE *f = fopen(filename,"r");
        int res;
        if(!f){
            h->info->error = "Failed open file for reading";
            return -1;
        }
        res = _gd_create_from_stream(h,f);
        fclose(f);
        return res;
    }
}


static cairo_surface_t *
_gd_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *visible_layers,int prefered_variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gd_get_mime_types,
        .new = _gd_new,
        .free = _gd_free,
        .create_from_data = _gd_create_from_data,
        .create_from_file = _gd_create_from_file,
        .get_image_surface = _gd_get_image_surface,
    };
    return &p;
}
