/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <assert.h>
#include <debug.h>
#include <libheif/heif.h>
#include <nil-cairo.h>
#include <stdlib.h>                     /* malloc(), free() */


typedef struct{
    int width;
    int height;
    cairo_surface_t *surface;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_heif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/heic",
        "image/heic-sequence",
        "image/heif",
        "image/heif-sequence",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_heif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_heif_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_img(abydos_plugin_handle_t *h, struct heif_image *img,int width, int height)
{
    const uint8_t *pixels;
    int stride;
    pixels = heif_image_get_plane_readonly(img, heif_channel_interleaved, &stride);
    return nil_cairo_surface_create_from_direct(pixels,width,height,stride,4,NILE_FORMAT_r8g8b8a8,0);
}


static int
_heif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    struct heif_error err;
    struct heif_context* hc;
    struct heif_image_handle* imgh = NULL;
    heif_item_id *id = NULL;
    struct heif_image* img = NULL;
    int res = -1;
    int i;

    hc = heif_context_alloc();

    err = heif_context_read_from_memory_without_copy(hc, bytes, len, NULL);
    if(err.code != heif_error_Ok) {
        h->info->error = err.message;
        goto cleanup;
    }

    h->info->page_count = heif_context_get_number_of_top_level_images(hc);
    DEBUG_p("images: %d",h->info->page_count);
    h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));
    id = (heif_item_id*)malloc(h->info->page_count * sizeof(heif_item_id));
    heif_context_get_list_of_top_level_image_IDs(hc, id, h->info->page_count);

    for(i=0;i<h->info->page_count;++i){
        int width;
        int height;
        err = heif_context_get_image_handle(hc, id[i], &imgh);
        if(err.code != heif_error_Ok) {
            h->info->error = err.message;
            goto cleanup;
        }

        err = heif_decode_image(imgh, &img, heif_colorspace_RGB, heif_chroma_interleaved_RGBA, NULL);
        if(err.code != heif_error_Ok) {
            h->info->error = err.message;
            goto cleanup;
        }
        heif_image_handle_release(imgh);
        imgh = NULL;

        h->page[i].width = width = heif_image_get_width(img, heif_channel_interleaved);
        h->page[i].height = height = heif_image_get_height(img, heif_channel_interleaved);
        h->page[i].surface = _surface_from_img(h,img,width,height);
        heif_image_release(img);
        img = NULL;
        if(width>h->info->width)
            h->info->width = width;
        if(height>h->info->height)
            h->info->height = height;
    }
    res = 0;

cleanup:
    if(img)
        heif_image_release(img);
    if(id)
        free(id);
    if(imgh)
        heif_image_handle_release(imgh);
    if(res<0)
        heif_context_free(hc);

    return res;
}


static cairo_surface_t *
_heif_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


static void
_heif_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    DEBUG_p("begin");
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _heif_get_mime_types,
        .new = _heif_new,
        .free = _heif_free,
        .create_from_data = _heif_create_from_data,
        .get_image_surface = _heif_get_image_surface,
        .get_duration = _heif_get_duration
    };
    return &p;
}
