/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <assert.h>
#include <string.h>

#include "buffer.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif

static void
_abydos_buffer_store(abydos_buffer_t *buf, const uint8_t *data, size_t len)
{
    assert(buf->buffer_len + len <= ABYDOS_BUFFER_SIZE);
    if(0 == buf->buffer_len){
        buf->buffer_pos = 0;
    }else if(buf->buffer_pos + buf->buffer_len + len > ABYDOS_BUFFER_SIZE){
        memmove(buf->buffer,buf->buffer+buf->buffer_pos,buf->buffer_len);
        buf->buffer_pos = 0;
    }
    memcpy(buf->buffer+buf->buffer_pos+buf->buffer_len,data,len);
    buf->buffer_len += len;
}


abydos_buffer_t *
abydos_buffer_clear(abydos_buffer_t * buf)
{
    buf->buffer_len = 0;
    buf->data_len = 0;
    return buf;
}


void
abydos_buffer_add(abydos_buffer_t *buf,const uint8_t *data, size_t len)
{
    if(buf->data_len > 0)
        abydos_buffer_privatize(buf);
    buf->data = data;
    buf->data_len = len;
    buf->data_pos = 0;
}


void
abydos_buffer_add_back(abydos_buffer_t * buf,const uint8_t *data,size_t len)
{
    if(buf->buffer_len > 0){
        assert(len <= buf->buffer_pos);
        buf->buffer_len += len;
        buf->buffer_pos -= len;
        memcpy(buf->buffer+buf->buffer_pos,data,len);
    }else{
        memcpy(buf->buffer,data,len);
        buf->buffer_len = len;
        buf->buffer_pos = 0;
    }
}


int
abydos_buffer_pop(abydos_buffer_t *buf)
{
    int b;
    assert(abydos_buffer_size(buf) >= 1);
    if(buf->buffer_len > 0){
        b = buf->buffer[buf->buffer_pos++];
        --buf->buffer_len;
    }else{
        b = buf->data[buf->data_pos++];
        --buf->data_len;
    }
    return b;
}



void
abydos_buffer_privatize(abydos_buffer_t *buf)
{
    _abydos_buffer_store(buf,buf->data+buf->data_pos,buf->data_len);
    buf->data_len = 0;
}


const uint8_t *
abydos_buffer_copy(abydos_buffer_t *buf,uint8_t *dst,size_t len)
{
    if(abydos_buffer_size(buf) < len)
        return NULL;
    if(buf->buffer_len > 0){
        int bytes = MIN(len,buf->buffer_len);
        memcpy(dst,buf->buffer+buf->buffer_pos,bytes);
        buf->buffer_pos += bytes;
        buf->buffer_len -= bytes;
        dst += bytes;
        len -= bytes;
    }
    if(len>0){
        memcpy(dst,buf->data+buf->data_pos,len);
        buf->data_pos += len;
        buf->data_len -= len;
        dst += len;
    }
    return dst;
}


const uint8_t *
abydos_buffer_get(abydos_buffer_t * buf,size_t len)
{
    const uint8_t *data;
    if(abydos_buffer_size(buf) < len)
        return NULL;
    if(0 == buf->buffer_len){
        data = buf->data + buf->data_pos;
        buf->data_len -= len;
        buf->data_pos += len;
    }else{
        if(len > buf->buffer_len){
            size_t left = len - buf->buffer_len;
            _abydos_buffer_store(buf,buf->data+buf->data_pos,left);
            buf->data_len -= left;
            buf->data_pos += left;
        }
        data = buf->buffer + buf->buffer_pos;
        buf->buffer_len -= len;
        buf->buffer_pos += len;
    }
    return data;
}


#if 0
const uint8_t *
abydos_buffer_peek(abydos_buffer_t *buf,size_t *r_size)
{
    if(buf->buffer_len > 0){
        *r_size = buf->buffer_len;
        return buf->buffer;
    }else if(buf->data_len > 0){
        *r_size = buf->data_len;
        return buf->data;
    }else{
        *r_size = 0;
        return NULL;
    }
}


int
abydos_buffer_peek_at(abydos_buffer_t *buf,int index)
{
    if(index < buf->buffer_len)
        return buf->buffer[buf->buffer_pos + index];
    index -= buf->buffer_len;
    if(index < buf->data_len)
        return buf->data[buf->data_pos + index];
    return -1;
}
#endif


void
abydos_buffer_skip(abydos_buffer_t * buf,size_t len)
{
    assert(len <= abydos_buffer_size(buf));
    if(len < buf->buffer_len){
        buf->buffer_len -= len;
        buf->buffer_pos += len;
    }else{
        len -= buf->buffer_len;
        buf->buffer_len = 0;
        buf->data_len -= len;
        buf->data_pos += len;
    }
}
