/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_kiss_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-kiss-cel",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_kiss_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_kiss_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16(a) le16toh(*(uint16_t*)(a))


static cairo_surface_t *
_surface_from_grey4(const uint8_t* src,int width,int height)
{
    int src_stride = (width-1)/2+1;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int pix = src[x>>1];
            if(x&1){
                pix &= 0x0f;
                pix |= pix<<4;
            }else{
                pix &= 0xf0;
                pix |= pix>>4;
            }
            dst[x] = pix ? pix<<16|pix<<8|pix<<0|255<<24 : 0;
        }
        src += src_stride;
        dst += dst_stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_grey8(const uint8_t* src,int width,int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    while(height-->0){
        int x;
        DEBUG_p("rows left: %d",height+1);
        for(x=0;x<width;++x){
            int pix = *src++;
            dst[x] = pix ? pix<<16|pix<<8|pix<<0|255<<24 : 0;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_rgba(const uint8_t* src,int width,int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    while(height-->0){
        int x;
        for(x=0;x<width;++x){
            int alpha = src[3];
            int r = alpha*src[2]/255;
            int g = alpha*src[1]/255;
            int b = alpha*src[0]/255;
            dst[x] = r<<16|g<<8|b<<0|alpha<<24;
            src += 4;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_kiss_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    int bpp;
    int needed;
    if(*(uint32_t*)data != *(uint32_t*)"KiSS"){
        h->info->error = "Invalid header";
        h->info->width = get16(data+0);
        h->info->height = get16(data+2);
        bpp = 4;
        data+=4;
        len-=4;
    }else{
#if 0
        if(data[4] != 0x20)
            h->info->error = "Not a bitmap";
            return -1;
        }
#endif
        bpp = data[5];
        if(bpp != 4 && bpp != 8 && bpp != 32){
            h->info->error = "Unknown file encoding";
            return -1;
        }
        h->info->width = get16(data+8);
        h->info->height = get16(data+10);
        DEBUG_p("offset: %dx%d",get16(data+12),get16(data+14));
        data += 32;
        len -= 32;
    }

    needed = bpp*h->info->width*h->info->height/8;


    DEBUG_p("dimensions: %dx%d",h->info->width,h->info->height);
    DEBUG_p("needed bytes: %d, bytes left: %d",needed,(int)len);

    if(needed > len){
        h->info->error = "The file is too short";
        return -1;
    }

    switch(bpp){
    case 4:
        h->surface = _surface_from_grey4((const uint8_t*)data,h->info->width,h->info->height);
        break;
    case 8:
        h->surface = _surface_from_grey8((const uint8_t*)data,h->info->width,h->info->height);
        break;
    case 32:
        h->surface = _surface_from_rgba((const uint8_t*)data,h->info->width,h->info->height);
        break;
    }
    return 0;
}



static cairo_surface_t *
_kiss_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _kiss_get_mime_types,
        .new = _kiss_new,
        .free = _kiss_free,
        .create_from_data = _kiss_create_from_data,
        .get_image_surface = _kiss_get_image_surface
    };
    return &p;
}
