/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <nil-mcga.h>
#include <stdlib.h>
#include <string.h>


#define get8(a) (*(uint8_t*)(a))


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_prf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-fastgraph-prf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_prf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_prf_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


typedef struct{
    const char *data;
    const char *end;
    int run;
    int value;
}rle_t;


static void
_rle_init(rle_t *rle, const char *data, size_t len)
{
    rle->data = data;
    rle->end = data + len;
    rle->run = 0;
}


static int
_rle_get(rle_t *rle)
{
    if(rle->run <= 0){
        if(rle->data + 2 > rle->end)
            return -1;
        rle->value = get8(rle->data++);
        rle->run = get8(rle->data++);
        DEBUG_p("run: %d, value: %d", rle->run, rle->value);
    }
    --rle->run;
    return rle->value;
}


static int
_prf_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const char *end = data + len;
    nile_t ni;
    rle_t rle;
    int res = -1;
    int i;

    if(memcmp(data, "F\0A\0S\0T\0G\0R\0A\0F\0", 16))
        return -1;
    data += 16;

    if(data + 10 > end)
        return -1;

    nile_init(&ni);
    ni.width = get8(data) + 256 * get8(data+2) + 1;
    ni.height = get8(data+4) + 256 * get8(data+6) + 1;
    data += 10;

    _rle_init(&rle, data, end-data);

    nile_ensure_indexed(&ni, ni.width, ni.height);
    for(i=0;i<ni.width*ni.height;++i){
        int b = _rle_get(&rle);
        if(b < 0)
            goto out;
        ni.pixels[i] = b;
    }
    ni.palette = (uint8_t*)nil_mcga_palette;
    ni.colors = 256;
    ni.bpc = 3;
    ni.format = NILE_FORMAT_r8g8b8;

    h->info->width = ni.width;
    h->info->height = ni.height;
    if(ni.height <= 400)
        h->info->pixel_ratio = 5.0/6;
#if 1
    h->surface = nil_cairo_surface_create(&ni, NILE_FLIPPED);
#else
    h->surface = nil_cairo_surface_create_from_indexed(ni.pixels,
                                                       ni.width,
                                                       ni.height,
                                                       nile_get_stride(&ni),
                                                       ni.palette,
                                                       ni.colors,
                                                       ni.bpc,
                                                       ni.format,
                                                       NILE_FLIPPED);
#endif
    res = 0;
out:
    nile_done(&ni);
    return res;
}


static cairo_surface_t *
_prf_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _prf_get_mime_types,
        .new = _prf_new,
        .free = _prf_free,
        .create_from_data = _prf_create_from_data,
        .get_image_surface = _prf_get_image_surface
    };
    return &p;
}
