/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

#if __BYTE_ORDER == __BIG_ENDIAN
# define ID(a,b) ((a<<8)|(b<<0))
#else
# define ID(a,b) ((a<<0)|(b<<8))
#endif

#define ID_P2 ID('P','2')
#define ID_CM ID('C','M')
#define ID_PM ID('P','M')
#define ID_PS ID('P','S')
#define ID_TS ID('T','S')

#define get8(a) (*(uint8_t*)(a))
#define get16(a) le16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_tp_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-telepaint",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_tp_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_tp_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


typedef struct{
    const char *data;
    const char *end;
    int copy;
    int run;
    int value;
    int offs;
}rle_t;


static void
_rle_init(rle_t *rle, const char *data, size_t len)
{
    rle->data = data + 2;
    rle->end = data + len;
    rle->copy = get16(data);
    rle->run = 0;
    rle->offs = 0;
}


static int
_rle_get(rle_t *rle)
{
    if(rle->copy <= 0 && rle->run <= 0){
        if(rle->data + 4 > rle->end)
            return -1;
        rle->offs += 4;
        rle->value = get8(rle->data);
        rle->run = get8(rle->data+1);
        rle->copy = get16(rle->data+2) - rle->offs;
        rle->data += 4;
        DEBUG_p("run: %d, copy: %d", rle->run, rle->copy);
    }
    if(rle->run > 0){
        --rle->run;
        return rle->value;
    }
    if(rle->data >= rle->end)
        return -1;
    ++rle->offs;
    --rle->copy;
    return get8(rle->data++);
}

/*

---DERBYVSL---version 4-----------------------------------------------------

TS:
                  _244_ _244_ _21__
                  F4 00 F4 00 15 00 00 00 01 08
                                          _0___
00 00 00 01 29 12 B8 10 98 10 7B 0B 75 0B 00 00
      _5___          _10__          _15__
E0 0A 05 00 EA E0 0A 0A 00 EA E0 0A 0F 00 EA E0
   _20__          _25__          _30__
0A 14 00 EA E0 0A 19 00 EA E0 0A 1E 00 EA E0 0A
_35__          _40__
23 00 EA E0 0A 28 00 EA E0 0A 2D 00 EA E0 0A 32

PS:
               _211_ _211_ _21__
               D3 00 D3 00 15 00 00 00 01 08 00
               _0___       _5___          _10__
00 00 01 D5 0C 00 00 E0 0A 05 00 EA E0 0A 0A 00
         _15__          _20__
EA E0 0A 0F 00 EA E0 0A 14 00 EA E0 0A 19 00 EA

E0 0A 1E 00 EA E0 0A 23 00 EA E0 0A 28 00 EA E0

0A 2D 00 EA E0 0A 32 00 EA E0 0A 37 00 EA E0 15

---FRACTAL1---version 2-----------------------------------------------------

PM:

__width__ _stride?_ __height_
0x40 0x01 0x40 0x01 0xc8 0x00 0x00 0x00 0x01 0x08 0x00 0x00 0x00 0x00 0x00 0x00

---MONKEY-----version 2-----------------------------------------------------
                                                            _=692____
PM:

__width__ _stride?_ __height_
0x40 0x01 0x40 0x01 0xc8 0x00 0x00 0x00 0x01 0x08 0x08 0x00 0x00 0x00 0x2a 0x2f

---MOUSE------version 4-----------------------------------------------------

PM:

__width__ _stride?_ __height_
0x40 0x01 0x40 0x01 0xc8 0x00 0x00 0x00 0x01 0x08 0x00 0x00 0x00 0x01 0x7a 0xd4
__=0_____      _=5_ __=16____
0x00 0x00 0x37 0x05 0x10 0x00 0x70 0x37 0x37 0x70 0x70 0x70 0x37 0x37 0x37 0x31

0x70 0x37 0x70 0x06 0x16 0x00 0x37 0x65 0x70 0x05 0x2f 0x00 0x37 0x37 0x37 0x31

0x31 0x70 0x70 0x70 0x37 0x70 0x70 0x37 0x65 0x70 0x70 0x37 0x37 0x70 0x37 0x37

---VASE2--------------------------------------------------------------------

PM:

__width__ _stride?_ __height_
0x40 0x01 0x40 0x01 0xc8 0x00 0x00 0x00 0x01 0x08 0x00 0x00 0x00 0x01 0x6b 0xcd
__=7_____                                         _=6_ __=34____
0x07 0x00 0xd9 0x92 0xd9 0xd9 0xd9 0x92 0xd9 0x92 0x06 0x22 0x00 0xd9 0xd9 0x92

0x92 0x92 0xa4 0xa4 0xa4 0x92 0xdb 0xdb 0x92 0x92 0x92 0x92 0xa4 0xdb 0xdb 0x8c
                         _=9_ =39_                _=9_ __=43____
0x8c 0xdb 0xdb 0xa4 0xdb 0x09 0x27 0x00 0x91 0xdb 0x09 0x2b 0x00 0x91 0x09 0x31


*/

static int
_tp_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const char *end = data + len;
    int version;
    const char *p;
    nile_t ni;
    int res = -1;

    if(data + 4 > end)
        return -1;
    if(*(uint16_t*)data != ID_P2)
        return -1;

    version = get16(data + 2);
    if(version != 2 && version != 4)
        return -1;
    data += 4;

    nile_init(&ni);
    ni.type = NILE_TYPE_INDEXED;
    ni.colors = 256;
    ni.format = NILE_FORMAT_r6g6b6;
    ni.bpc = 3;

    for(p=data;;p+=6){
        int id;
        const char *addr;
        if(p + 2 > end)
            return -1;
        id = *(uint16_t*)p;
        if(!id)
            break;
        if(p + 6 > end)
            return -1;
        addr = data + get32(p + 2);
        switch(id){
        case ID_CM:
            if(addr + 768 > end)
                goto out;
            ni.palette = (uint8_t*)addr;
            break;
        case ID_PM:
            if(4 == version){
                rle_t rle;
                int i;
                if(addr + 16 > end)
                    goto out;
                ni.width = get16(addr);
                ni.height = get16(addr+4);
                addr += 16;

                nile_ensure_indexed(&ni, ni.width, ni.height);
                _rle_init(&rle, addr, end-addr);
                for(i=0;i<ni.width*ni.height;++i){
                    int b = _rle_get(&rle);
                    if(b < 0)
                        goto out;
                    ni.pixels[i] = b;
                }
            }else{
                if(addr + 14 > end)
                    goto out;
                ni.width = get16(addr);
                ni.height = get16(addr+4);
                addr += 14;
                if(addr + ni.width * ni.height > end)
                    goto out;
                ni.pixels = (uint8_t*)addr;
            }
        case ID_PS:
            break;
        case ID_TS:
            break;
        default:
            DEBUG_p("unknown entry: %.2s", data);
            break;
        }
    }

    if(!ni.palette){
        DEBUG_p("colormap not found");
        goto out;
    }
    if(!ni.pixels){
        DEBUG_p("pixmap not found");
        goto out;
    }

    h->info->width = ni.width;
    h->info->height = ni.height;
    h->info->pixel_ratio = 5.0/6;
    h->surface = nil_cairo_surface_create(&ni, 0);
    res = 0;
out:
    nile_done(&ni);
    return res;
}


static cairo_surface_t *
_tp_get_image_surface(abydos_plugin_handle_t *h,
                      int page,
                      cairo_bool_t *layer_visibility,
                      int variant,
                      double time,
                      int frame)
{
    return cairo_surface_reference(h->surface);
}


static void
_tp_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _tp_get_mime_types,
        .new = _tp_new,
        .free = _tp_free,
        .create_from_data = _tp_create_from_data,
        .get_image_surface = _tp_get_image_surface,
        .get_duration = _tp_get_duration
    };
    return &p;
}
