/*
 * Copyright (C) 2019 - 2022  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


typedef struct{
    cairo_surface_t *surface;
    int width;
    int height;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_ticalc_get_mime_types()
{
    static const char *mime_types[] = {
        "application/x-ti-variable",
        "application/x-tilp",
        "application/x-tilp-image",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ticalc_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_ticalc_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            if(h->page[i].surface)
                cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) le16toh(*(uint16_t*)(a))
#define get16be(a) be16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))


static cairo_surface_t *
_surface_from_bits(const uint8_t *src,int src_stride,int width,int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    uint16_t *dst = (uint16_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint16_t);
    while(height-->0){
        int x;
        for(x=0;x<src_stride;++x){
            int pix = src[x];
            int b;
            for(b=0;b<8 && x*8+b<width;++b){
                pix <<= 1;
                dst[x*8+b] = pix&256 ? 0 : 0xffffff;
            }
        }
        src += src_stride;
        dst += dst_stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_grayscale(const uint8_t *src,int src_stride,int width,int height)
{
#if 0
    /*
      Then converted to 24bits this results in colors that are not entirely
      gray, causing less efficiant PNG compression.
    */
    static const uint16_t palette[16] = {
        0x0000,0x1082,0x2104,0x3186,
        0x4228,0x52aa,0x632c,0x73ae,
        0x8c51,0x9cd3,0xad55,0xbdd7,
        0xce79,0xdefb,0xef7d,0xffff
    };
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    uint16_t *dst = (uint16_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint16_t);
#else
    static const uint32_t palette[16] = {
        0x00000000,0x00111111,0x00222222,0x00333333,
        0x00444444,0x00555555,0x00666666,0x00777777,
        0x00888888,0x00999999,0x00aaaaaa,0x00bbbbbb,
        0x00cccccc,0x00dddddd,0x00eeeeee,0x00ffffff
    };
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
#endif
    while(height-->0){
        int x;
        for(x=0;x<src_stride;++x){
            int pix = src[x];
            int b;
            for(b=0;b<2 && x*2+b<width;++b){
                pix <<= 4;
                dst[x*2+b] = palette[(pix>>8)&15];
            }
        }
        src += src_stride;
        dst += dst_stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_rgb565(const uint16_t *src,int src_stride,int width,int height)
{
#if 1
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    uint16_t *dst = (uint16_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint16_t);
    dst += height*dst_stride;
    while(height-->0){
        int x;
        dst -= dst_stride;
        for(x=0;x<width;++x)
            dst[x] = le16toh(src[x]);
        src += src_stride;
    }
#else
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    dst += height*dst_stride;
    while(height-->0){
        int x;
        dst -= dst_stride;
        for(x=0;x<width;++x){
            int pix = get16(src+x);
            int r = pix>>11;
            int g = (pix>>5)&63;
            int b = pix&31;
            r = r<<3|r>>2;
            g = g<<2|g>>4;
            b = b<<3|b>>2;
            dst[x] = r<<16|
                     g<< 8|
                     b<< 0;
        }
        src += src_stride;
    }
#endif
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_handle_ti8x(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    off_t offs;
    int page;

    data += 55;
    len -= 55;

    /* scan for number of pages */
    h->info->page_count = 0;
    offs = 0;
    while(offs+15 < len){
        /* data[offs] should be 11 */
        int size = get16(data+offs+2);
        int type = get8(data+offs+4);
        if(offs+15+size<=len && (7==type||26==type))
            ++h->info->page_count;
        offs += 15+size;
    }
    DEBUG_p("found %d pages",h->info->page_count);
    if(h->info->page_count<=0){
        h->info->error = "No image variables found";
        return -1;
    }

    h->page = (page_t*)malloc(h->info->page_count*sizeof(page_t));

    page = 0;
    while(len>15){
        int size = get16(data+2);
        DEBUG_p("size: %d",size);
        if(15+size>len)
            break;
        switch(get8(data+4)){
        case 7: /* image */
            if(21947==size){
                h->page[page].surface = _surface_from_grayscale((const uint8_t*)data+17+2,133,265,165);
                h->page[page].width = 265;
                h->page[page].height = 165;
            }else{
                h->page[page].surface = _surface_from_bits((const uint8_t*)data+17,12,95,63);
                h->page[page].width = 95;
                h->page[page].height = 63;
            }
            if(h->page[page].width > h->info->width)
                h->info->width = h->page[page].width;
            if(h->page[page].height > h->info->height)
                h->info->height = h->page[page].height;
            ++page;
            break;
        case 26: /* background */
            h->page[page].surface = _surface_from_rgb565((const uint16_t*)(data+17+3),134,133,83);
            h->page[page].width = 133;
            h->page[page].height = 83;
            if(h->page[page].width > h->info->width)
                h->info->width = h->page[page].width;
            if(h->page[page].height > h->info->height)
                h->info->height = h->page[page].height;
            ++page;
            break;
        }
        data += 15+size;
        len -= 15+size;
    }
    return 0;
}


static int
_handle_ti9x_image(page_t *page,const char *data,size_t len)
{
    //int size;
    int stride;
    if(len<10)
        return -1;
    //size = get16be(data+4);
    page->width = get16be(data+8);
    page->height = get16be(data+6);
    stride = ((page->width-1)/8)+1;
    if(10+stride*page->height>len)
        return -1;
    page->surface = _surface_from_bits((const uint8_t*)data+10,stride,page->width,page->height);
    return 0;
}


static int
_handle_ti9x(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    int count;
    int i;

    if(len<82){
        h->info->error = "The file is too small";
        return -1;
    }

    count = get16(data+58);
    DEBUG_p("count: %d",count);
    h->info->page_count = 0;
    for(i=0;i<count;++i){
        if(16==get16(data+60+i*16+12))
            ++h->info->page_count;
    }
    if(h->info->page_count<1){
        h->info->error = "The file contains no image";
        return -1;
    }
    h->page = (page_t*)malloc(h->info->page_count*sizeof(page_t));
    h->info->page_count = 0;
    for(i=0;i<count;++i){
        if(16==get16(data+60+i*16+12)){
            page_t *page = &h->page[h->info->page_count++];
            int offs = get16(data+60+i*16);
            if(_handle_ti9x_image(page,data+offs,len-offs)<0){
                page->surface = NULL;
            }else{
                if(page->width>h->info->width)
                    h->info->width = page->width;
                if(page->height>h->info->height)
                    h->info->height = page->height;
            }
        }
    }
    return 0;
}


static int
_ticalc_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    switch(get16(data+8)){
    case 0x0a1a:
        return _handle_ti8x(h,data,len);
    case 0x0001:
        return _handle_ti9x(h,data,len);
    default:
        DEBUG_p("unhandled signature: %d",get16(data+8));
        return -1;
    }
}


static cairo_surface_t *
_ticalc_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ticalc_get_mime_types,
        .new = _ticalc_new,
        .free = _ticalc_free,
        .create_from_data = _ticalc_create_from_data,
        .get_image_surface = _ticalc_get_image_surface
    };
    return &p;
}
