/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#define MIPTEX_VARIANT_COUNT 4

#define SETMAX(a,b) if((b)>(a))(a)=(b)

typedef struct{
    uint8_t magic[4];
    int32_t numentries;
    int32_t diroffset;
}wadhead_t;

typedef struct{
    uint32_t offset;
    uint32_t dsize;
    char name[8];
}wad1entry_t;

typedef struct{
    uint32_t offset;
    uint32_t dsize;
    uint32_t msize;
    uint8_t type;
    uint8_t cmprs;
    uint16_t dummy;
    char name[16];
}wad2entry_t;

typedef struct{
    uint32_t width;
    uint32_t height;
}pichead_t;

typedef struct{
    char name[16];
    uint32_t width;
    uint32_t height;
    uint32_t offset[4];
}miptexhead_t;

typedef struct{
    int width;
    int height;
    cairo_surface_t *surface;
}variant_t;

typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

typedef struct{
    int width;
    int height;
    enum{
        T_PICTURE,
        T_MIPTEX
    }type;
    union{
        cairo_surface_t *surface;
        struct{
            variant_t variant[MIPTEX_VARIANT_COUNT];
        };
        struct{
            frame_t *frame;
            int frame_count;
        };
    };
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    uint32_t palette[256];
    page_t *page;
};


static const char **
_wad_get_mime_types()
{
    static const char *mime_types[] = {
        "application/x-doom-wad",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_wad_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_wad_free(abydos_plugin_handle_t *h)
{
    free(h);
}


#define get32(a) le32toh(*(uint32_t*)(a))


static int
_convert_palette(const uint8_t *src_palette, uint32_t *cairo_palette)
{
    int i;
    for(i=0;i<256;++i){
        int r = src_palette[i*3+0];
        int g = src_palette[i*3+1];
        int b = src_palette[i*3+2];
#if 0
        r = r<<2 | r>>4;
        g = g<<2 | g>>4;
        b = b<<2 | b>>4;
#endif
        cairo_palette[i] = r<<16 | g<<8 | b;
    }
    return 0;
}


static cairo_surface_t *
_surface_from_data(const uint8_t *src, int width, int height, const uint32_t *palette)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    while(height-->0){
        int x;
        for(x=0;x<width;++x)
            dst[x] = palette[*src++];
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


#if 0
static cairo_surface_t *
_surface_from_picture(const uint8_t *src, int width, int height, const uint32_t *palette)
{
    return _surface_from_data(src, width, height, palette);
}
#endif


static void
_adjust_max_size(abydos_plugin_handle_t* h, page_t *page)
{
    SETMAX(page->width,page->variant[0].width);
    SETMAX(page->height,page->variant[0].height);
}


static int
_handle_picture(abydos_plugin_handle_t* h, page_t *page, const uint8_t *data, size_t len)
{
    //pichead_t *head = (pichead_t*)data;
    page->type = T_PICTURE;
    page->width = get32(data + offsetof(pichead_t,width));
    page->height = get32(data + offsetof(pichead_t,height));
    /* FIXME!! check sanity of image dimensions */
    DEBUG_p("picture: %dx%d",page->width,page->height);
    data += sizeof(pichead_t);
    len -= sizeof(pichead_t);
    page->surface = _surface_from_data(data,page->width,page->height,h->palette);
    return 0;
}


static int
_handle_miptex(abydos_plugin_handle_t *h, page_t *page, const uint8_t *data, size_t len)
{
    //miptexhead_t *head = (miptexhead_t*)data;
    int width = get32(data + offsetof(miptexhead_t,width));
    int height = get32(data + offsetof(miptexhead_t,height));
    int i;
    DEBUG_p("begin");

    page->type = T_MIPTEX;

    for(i=0;i<MIPTEX_VARIANT_COUNT;++i){
        int size = width*height;
        int offset = get32(data + offsetof(miptexhead_t,offset) + i*4);
        if(offset+size > len)
            return -1;
        page->variant[i].width = width;
        page->variant[i].height = height;
        page->variant[i].surface = _surface_from_data(data+offset,width,height,h->palette);
        width /= 2;
        height /= 2;
    }
    _adjust_max_size(h,page);
    return 0;
}


static int
_handle_wad1(abydos_plugin_handle_t *h,const uint8_t *data, size_t len)
{
    DEBUG_p("begin");
    return -1;
}


static int
_handle_wad2(abydos_plugin_handle_t *h,const uint8_t *data, size_t len, int wad3)
{
    //wadhead_t *head = (wadhead_t*)data;
    int numentries;
    int diroffset;
    page_t *page;
    int i;
    DEBUG_p("begin");

    numentries = get32(data + offsetof(wadhead_t,numentries));
    diroffset = get32(data + offsetof(wadhead_t,diroffset));

    DEBUG_p("entries: %d",numentries);
    DEBUG_p("offset : %d",diroffset);

    if(len < diroffset + numentries * sizeof(wad2entry_t)){
        /* file to short */
        return -1;
    }

    /* Generate a default palette */
    for(i=0;i<256;++i)
        h->palette[i] = i<<16|i<<8|i;

    /* Count the pages */
    h->info->page_count = 0;
    for(i=0;i<numentries;++i){
        wad2entry_t *entry = (wad2entry_t*)&data[diroffset + i * sizeof(wad2entry_t)];
        switch(entry->type){
        case 'B':
        case 'C':
        case 'D':
            ++h->info->page_count;
            break;
        }
    }
    page = h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));

    for(i=0;i<numentries;++i){
        const uint8_t *entry = &data[diroffset + i * sizeof(wad2entry_t)];
        int offset = get32(entry + offsetof(wad2entry_t,offset));
        int dsize = get32(entry + offsetof(wad2entry_t,dsize));
        DEBUG_p("entry %d offset: %d",i,offset);
        DEBUG_p("entry %d size  : %d",i,dsize);
        DEBUG_p("entry %d name  : %16s",i,entry+offsetof(wad2entry_t,name));
        /* FIXME!! check that the offset and size is valid */
        switch(entry[offsetof(wad2entry_t,type)]){
        case '@':
            DEBUG_p("palette");
            if(dsize >= 256*3)
                _convert_palette(data+offset,h->palette);
            break;
        case 'B':
            _handle_picture(h,page++,data+offset,dsize);
            break;
        case 'C':
        case 'D':
            DEBUG_p("miptex");
            _handle_miptex(h,page++,data+offset,dsize);
            break;
        case 'E':
            DEBUG_p("flat picture");
            break;
        default:
            DEBUG_p("unknown: %c",entry[offsetof(wad2entry_t,type)]);
            break;
        }
    }

    return 0;
}


static int
_wad_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    DEBUG_p("begin");
    if(len<sizeof(wadhead_t)){
        /* file too short */
        return -1;
    }
    if(!memcmp(data,"IWAD",4) || !memcmp(data,"PWAD",4))
        return _handle_wad1(h,(const uint8_t*)data,len);
    if(!memcmp(data,"WAD2",4))
        return _handle_wad2(h,(const uint8_t*)data,len,0);
    if(!memcmp(data,"WAD3",4))
        return _handle_wad2(h,(const uint8_t*)data,len,1);
    /* unknown file type */
    return -1;
}


static cairo_surface_t *
_wad_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    switch(h->page[page].type){
    case T_PICTURE:
        return cairo_surface_reference(h->page[page].surface);
    case T_MIPTEX:
        return cairo_surface_reference(h->page[page].variant[variant].surface);
    }
    /* not reached */
    return NULL;
}


static void
_wad_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->width = h->page[page].width;
    info->height = h->page[page].height;
    switch(h->page[page].type){
    case T_PICTURE:
        break;
    case T_MIPTEX:
        info->variant_count = MIPTEX_VARIANT_COUNT;
        break;
    }
}


static void
_wad_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size)
{
    int i;
    switch(h->page[page].type){
    case T_PICTURE:
        /* this should not be called */
        size[0].width = h->page[page].width;
        size[0].height = h->page[page].height;
        break;
    case T_MIPTEX:
        for(i=0;i<MIPTEX_VARIANT_COUNT;++i){
            size[i].width = h->page[page].variant[i].width;
            size[i].height = h->page[page].variant[i].height;
        }
        break;
    }
}


static void
_wad_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _wad_get_mime_types,
        .new = _wad_new,
        .free = _wad_free,
        .create_from_data = _wad_create_from_data,
        .get_image_surface = _wad_get_image_surface,
        .get_page_info = _wad_get_page_info,
        .get_variant_size = _wad_get_variant_size,
        .get_duration = _wad_get_duration
    };
    return &p;
}
