/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcpy() */

#include "loa-abstract-abydos.h"


static loa_abstract_t *
_create()
{
    loa_abstract_t *abst = malloc(sizeof(loa_abstract_t));
    abst->refcount = 1;
    return abst;
}


static cairo_surface_t *
_dup_surface(cairo_surface_t *src)
{
    int stride = cairo_image_surface_get_stride(src);
    int height = cairo_image_surface_get_height(src);
    cairo_surface_t *dst = cairo_image_surface_create(cairo_image_surface_get_format(src),
                                                      cairo_image_surface_get_width(src),
                                                      height);
    memcpy(cairo_image_surface_get_data(dst),
           cairo_image_surface_get_data(src),
           stride * height);
    cairo_surface_mark_dirty(dst);
    return dst;
}


static void
_ensure_surface(loa_abstract_t *abst)
{
    cairo_surface_t *surface;
    switch(abst->type){
    case ABST_EMPTY:
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, abst->width, abst->height);
        break;
    case ABST_ABYDOS:
        surface = abydos_get_image_surface(abst->ar);
        abydos_destroy(abst->ar);
        break;
    case ABST_PATTERN:
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, abst->width, abst->height);
        {
            cairo_t *cr = cairo_create(surface);
            cairo_set_source(cr, abst->pattern);
            cairo_paint(cr);
            cairo_destroy(cr);
        }
        cairo_pattern_destroy(abst->pattern);
        break;
    case ABST_SURFACE:
        return;
    }
    abst->type = ABST_SURFACE;
    abst->surface = surface;
}


loa_abstract_t *loa_abstract_abydos_create(void *userdata, int width, int height)
{
    loa_abstract_t *abst = _create();
    abst->type = ABST_EMPTY;
    abst->width = width;
    abst->height = height;
    return abst;
}


loa_abstract_t *loa_abstract_abydos_from_abydos(abydos_t *ar)
{
    loa_abstract_t *abst = _create();
    abst->type = ABST_ABYDOS;
    abydos_get_size(ar, &abst->width, &abst->height);
    abst->ar = ar;
    return abst;
}


void
loa_abstract_abydos_render(double x,
                           double y,
                           double xscale,
                           double yscale,
                           int flags,
                           loa_abstract_t *abst,
                           nil_clip_t *clip,
                           cairo_t *cr)
{
    if(ABST_EMPTY == abst->type)
        return;

    if(flags & LOA_ABSTRACT_REPLACE){
        cairo_rectangle(cr, x, y, abst->width * xscale, abst->height * yscale);
        cairo_clip(cr);
        cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    }else if(flags & LOA_ABSTRACT_UNDER){
        cairo_set_operator(cr, CAIRO_OPERATOR_DEST_OVER);
    }else if(flags & LOA_ABSTRACT_ADD){
        cairo_set_operator(cr, CAIRO_OPERATOR_ADD);
    }
    if(flags & LOA_ABSTRACT_FLIP_HORIZONTAL){
        x += abst->width * xscale;
        xscale *= -1;
    }
    if(flags & LOA_ABSTRACT_FLIP_VERTICAL){
        y += abst->height * yscale;
        yscale *= -1;
    }

    cairo_save(cr);
    if(clip){
        cairo_rectangle(cr, clip->x, clip->y, clip->width, clip->height);
        cairo_clip(cr);
    }
    cairo_translate(cr, x, y);
    cairo_scale(cr, xscale, yscale);
    if(ABST_ABYDOS == abst->type && !(flags & LOA_ABSTRACT_TILE)){
        abydos_render(abst->ar, cr);
    }else{
        cairo_pattern_t *pattern;
        switch(abst->type){
        case ABST_EMPTY:
            /* not reached */
            break;
        case ABST_ABYDOS:
            {
                cairo_surface_t *surface = abydos_get_image_surface(abst->ar);
                pattern = cairo_pattern_create_for_surface(surface);
                cairo_surface_destroy(surface);
            }
            break;
        case ABST_PATTERN:
            cairo_rectangle(cr, 0, 0, abst->width, abst->height);
            cairo_clip(cr);
            pattern = cairo_pattern_reference(abst->pattern);
            break;
        case ABST_SURFACE:
            pattern = cairo_pattern_create_for_surface(abst->surface);
            break;
        }
        cairo_pattern_set_filter(pattern, CAIRO_FILTER_NEAREST);
        if(flags & LOA_ABSTRACT_TILE)
            cairo_pattern_set_extend(pattern, CAIRO_EXTEND_REPEAT);
        cairo_set_source(cr, pattern);
        cairo_pattern_destroy(pattern);
        cairo_paint(cr);
    }
    cairo_restore(cr);
}


void
loa_abstract_abydos_free(void *userdata, loa_abstract_t *abst)
{
    if(--abst->refcount != 0)
        return;
    switch(abst->type){
    case ABST_EMPTY:
        break;
    case ABST_ABYDOS:
        abydos_destroy(abst->ar);
        break;
    case ABST_PATTERN:
        cairo_pattern_destroy(abst->pattern);
        break;
    case ABST_SURFACE:
        cairo_surface_destroy(abst->surface);
        break;
    }
    free(abst);
}


void
loa_abstract_abydos_get_size(void *userdata, loa_abstract_t *abst, int *width, int *height)
{
    *width = abst->width;
    *height = abst->height;
}


loa_abstract_t *
loa_abstract_abydos_copy(void *userdata, loa_abstract_t *src, int deep)
{
    if(deep){
        loa_abstract_t *dst = _create();
        dst->type = src->type;
        dst->width = src->width;
        dst->height = src->height;
        switch(src->type){
        case ABST_EMPTY:
            break;
        case ABST_ABYDOS:
            /* no need to do a deep copy yet, it will happen on write. */
            dst->ar = abydos_reference(src->ar);
            break;
        case ABST_PATTERN:
            /* no need to do a deep copy yet, it will happen on write. */
            dst->pattern = cairo_pattern_reference(src->pattern);
            break;
        case ABST_SURFACE:
            dst->surface = _dup_surface(src->surface);
            break;
        }
        return dst;
    }else{
        ++src->refcount;
        return src;
    }
}


void
loa_abstract_abydos_fill(void *userdata, loa_abstract_t *abst, double r, double g, double b, double a, nil_clip_t *clip)
{
    cairo_t *cr;
    if(ABST_EMPTY == abst->type && (!clip || (clip->x <= 0 && clip->y <= 0 && clip->width >= abst->width && clip->height >= abst->height))){
        abst->type = ABST_PATTERN;
        abst->pattern = cairo_pattern_create_rgba(r, g, b, a);
        return;
    }
    _ensure_surface(abst);
    cr = cairo_create(abst->surface);
    if(clip){
        cairo_rectangle(cr, clip->x, clip->y, clip->width, clip->height);
        cairo_clip(cr);
    }
    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    cairo_set_source_rgba(cr, r, g, b, a);
    cairo_paint(cr);
    cairo_destroy(cr);
}


void
loa_abstract_abydos_paste(void *userdata,
                          loa_abstract_t *dst,
                          double x,
                          double y,
                          double xscale,
                          double yscale,
                          int flags,
                          loa_abstract_t *src,
                          nil_clip_t *clip)
{
    cairo_t *cr;
    if(ABST_EMPTY == src->type)
        return;
    _ensure_surface(dst);
    cr = cairo_create(dst->surface);
    loa_abstract_abydos_render(x, y, xscale, yscale, flags, src, clip, cr);
    cairo_destroy(cr);
}
