/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


typedef enum{
    E_LITTLE = 1,
    E_BIG = 2
}endian_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t **surface;
    int count;
};


static const char **
_mpo_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-mpo",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_mpo_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_mpo_free(abydos_plugin_handle_t *h)
{
    if(h->surface){
        int i;
        for(i=0;i<h->count;++i)
            if(h->surface[i])
                cairo_surface_destroy(h->surface[i]);
        free(h->surface);
    }
    free(h);
}


#define get16le(a) le16toh(*(uint16_t*)(a))
#define get16be(a) be16toh(*(uint16_t*)(a))
#define get32le(a) le32toh(*(uint32_t*)(a))
#define get32be(a) be32toh(*(uint32_t*)(a))


static int
get16(const uint8_t *p, endian_t endian)
{
    switch(endian){
    case E_LITTLE: return get16le(p);
    case E_BIG: return get16be(p);
    }
    return -1;
}


static int
get32(const uint8_t *p, endian_t endian)
{
    switch(endian){
    case E_LITTLE: return get32le(p);
    case E_BIG: return get32be(p);
    }
    return -1;
}


static int
_handle_mpf(const uint8_t *p, off_t **offs)
{
    int entry_count;
    off_t offs_table = -1;
    int offs_count = 0;
    endian_t endian;
    const uint8_t *base;
    uint32_t size;
    int i;

    if('I'==p[0] && 'I'==p[1] && '*'==p[2] && '\0'==p[3]){
        endian = E_LITTLE;
    }else if('M'==p[0] && 'M'==p[1] && '\0'==p[2] && '*'==p[3]){
        endian = E_BIG;
    }else{
        DEBUG_p("unknown endian");
        /* Default to single picture. */
        return 1;
    }
    p += 4;
    base = p;

    size = get32(p, endian);
    p += (size - 4);

    entry_count = get16(p, endian);
    p += 2;

    for(i=0; i < entry_count; ++i){
        switch(get16(p, endian)){
        case 0xB001:
            p += 8;
            offs_count = get32(p, endian);
            p += 4;
            break;
        case 0xB002:
            p += 8;
            offs_table = get32(p, endian);
            p += 4;
            break;
        default:
            p += 12;
        }
    }

    if(offs_table > -1 && offs_count > 0){
        *offs = (off_t*)malloc(offs_count*sizeof(off_t));
        p = base + offs_table;
        (*offs)[0] = 0;
        for(i=1; i < offs_count; ++i){
            (*offs)[i] = get32(p, endian);
            p += 16;
        }
    }

    return offs_count;
}


static int
_get_offsets(const uint8_t *p, const uint8_t *end, off_t **offs)
{
    /* skip header */
    p += 2;

    for(;;){
        int size;
        if(p[0] != 0xff || p[1]<0xe0 || p[1]>0xe9)
            break;
        while((p[2] == 0xFF) && (p[3] == 0xFE))
            p += 2;

        size = get16be(p+2);

        switch(p[1]){
        case 0xe1:
            //abydos_load_feed(ar, (char*)p, 2+size);
            break;
        case 0xe2:
            if('M'==p[4] && 'P'==p[5] && 'F'==p[6] && '\0'==p[7]){
                /* Multi Picture Format */
                return _handle_mpf(p+8, offs);
            }
            break;
        }
        p += 2+size;
    }
    /* Default to single picture. */
    return 1;
}


cairo_surface_t *
_surface_from_jpeg(const uint8_t *p, const uint8_t *end)
{
    abydos_t *ar = abydos_create("image/jpeg");
    cairo_surface_t *surface;
    if(abydos_from_data(ar, (char*)p, end-p) < 0){
        surface = NULL;
    }else{
        surface = abydos_get_image_surface(ar);
    }
    abydos_destroy(ar);
    return surface;
}


static int
_mpo_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const uint8_t *p = (uint8_t*)data;
    const uint8_t *end = p+len;
    off_t *offs = NULL;
    int i;

    if((p[0] != 0xFF) || (p[1] != 0xD8)){
        DEBUG_p("not valid JPEG image");
        h->info->error = "Invalid JPEG data";
        return -1;
    }

    h->count = _get_offsets(p, end, &offs);
    h->surface = (cairo_surface_t**)malloc(h->count*sizeof(cairo_surface_t*));

    /* load the first picture separately since offs may not be allocated for
     * single picture.
     */
    h->surface[0] = _surface_from_jpeg((uint8_t*)data, end);
    for(i=1;i<h->count;++i)
        h->surface[i] = _surface_from_jpeg((uint8_t*)data + offs[i], end);

    if(offs)
        free(offs);

    return 0;
}


static cairo_surface_t *
_mpo_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface[h->count-1]);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _mpo_get_mime_types,
        .new = _mpo_new,
        .free = _mpo_free,
        .create_from_data = _mpo_create_from_data,
        .get_image_surface = _mpo_get_image_surface
    };
    return &p;
}
