/*
 * Copyright (C) 2018 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <glib.h>
#include <poppler.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    PopplerDocument *document;
};


static char *
_uri_from_filename(const char *filename){
    char *uri;
    char *cwd;
    char *path;
    if(g_path_is_absolute(filename))
        return g_filename_to_uri (filename, NULL,NULL);
    cwd = g_get_current_dir();
    path = g_build_filename(cwd,filename,NULL);
    g_free(cwd);
    uri = g_filename_to_uri (path,NULL,NULL);
    g_free(path);
    return uri;
}


static const char **
_pdf_get_mime_types()
{
    static const char *mime_types[] = {
        "application/pdf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_pdf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->scalable_size = 1;
    info->threadsafe = 1;
    h->document = NULL;
    return h;
}


static void
_pdf_free(abydos_plugin_handle_t *h)
{
    if(h->document)
        g_object_unref(h->document);
    free(h);
}


static int
_pdf_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    char* uri = _uri_from_filename(filename);

    h->document = poppler_document_new_from_file(uri,NULL,NULL);
    g_free(uri);
    if(!h->document){
        /* FIXME!! set info->error */
        return -1;
    }

    h->info->page_count = poppler_document_get_n_pages(h->document);

    return 0;
}


static void
_pdf_render(abydos_plugin_handle_t *h, cairo_t *cr, int page, int *visible_layers, int variant, double time, int frame)
{
    PopplerPage *pp = poppler_document_get_page(h->document,page);
    poppler_page_render((PopplerPage*)pp,cr);
    g_object_unref(pp);
}


static void
_pdf_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    PopplerPage *pp = poppler_document_get_page(h->document, page);
    double width;
    double height;
    poppler_page_get_size(pp, &width, &height);
    g_object_unref(pp);
    info->width = width;
    info->height = height;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _pdf_get_mime_types,
        .new = _pdf_new,
        .free = _pdf_free,
        .create_from_file = _pdf_create_from_file,
        .render = _pdf_render,
        .get_page_info = _pdf_get_page_info
    };
    return &p;
}
