/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
extern "C"{
#define new new_
#include <abydos-plugin.h>
#undef new
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>
}
#include "libraw/libraw.h"

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static cairo_surface_t *
_surface_from_image(libraw_processed_image_t *image)
{
    switch(image->type){
    case LIBRAW_IMAGE_BITMAP:
        return nil_cairo_surface_create_from_direct(image->data,
                                                    image->width,
                                                    image->height,
                                                    -1,
                                                    image->colors,
                                                    NILE_FORMAT_r8g8b8,
                                                    0);
    case LIBRAW_IMAGE_JPEG:
        /* FIXME!! */
        break;
    }
    return NULL;
}


static int
_convert(abydos_plugin_handle_t *h,LibRaw &proc)
{
    int res = -1;
    int ret;

    ret = proc.unpack();
    if(ret != LIBRAW_SUCCESS){
        DEBUG_p("unpack() failed: %s",libraw_strerror(ret));
        return -1;
    }
#if 0
    /* Unpack the thumbnail. */
    ret = proc.unpack_thumb();
    if(ret != LIBRAW_SUCCESS){
        DEBUG_p("unpack_thumb() failed: %s",libraw_strerror(ret));
        if(LIBRAW_FATAL_ERROR(ret))
            return -1;
    }else{
        ret = proc.dcraw_thumb_writer(thumbfn);
        if(ret != LIBRAW_SUCCESS){
            DEBUG_p("cannot write: %s",libraw_strerror(ret));
            /* In the case of fatal error, we should terminate processing of the current file. */
            if(LIBRAW_FATAL_ERROR(ret))
                return -1;
        }
    }
#endif
    ret = proc.dcraw_process();
    if(ret != LIBRAW_SUCCESS){
        DEBUG_p("dcraw_process() failed: %s",libraw_strerror(ret));
        return -1;
    }

    libraw_processed_image_t *image = proc.dcraw_make_mem_image();
    if(!image)
        return -1;
    DEBUG_p("channels: %d, bpp: %d",image->colors,image->bits);
    h->surface = _surface_from_image(image);
    if(h->surface){
        h->info->width = image->width;
        h->info->height = image->height;
        res = 0;
    }
    proc.dcraw_clear_mem(image);

    return res;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "image/x-sony-arw",
                "image/x-canon-cr2",
                "image/x-canon-crw",
                "image/x-kodak-dcr",
                "image/x-adobe-dng",
                "image/x-epson-erf",
                "image/x-kodak-k25",
                "image/x-kodak-kdc", "image/x-kdc",
                "image/x-minolta-mrw",
                "image/x-nikon-nef",
                "image/x-olympus-orf",
                "image/x-pentax-pef",
                "image/x-fuji-raf",
                "image/x-panasonic-raw",
                "image/x-sony-sr2",
                "image/x-sony-srf",
                "image/x-sigma-x3f",
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
            h->info = info;
            info->threadsafe = 1;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->surface)
                cairo_surface_destroy(h->surface);
            free(h);
        },
        .create_from_data = [](abydos_plugin_handle_t *h,const char *data, size_t len){
            LibRaw proc;
            int ret = proc.open_buffer((void*)data,len);
            if(ret != LIBRAW_SUCCESS){
                DEBUG_p("open_buffer() failed: %s",libraw_strerror(ret));
                return -1;
            }
            return _convert(h,proc);
        },
        .create_from_file = [](abydos_plugin_handle_t *h, const char *filename){
            LibRaw proc;
            int ret = proc.open_file(filename);
            if(ret != LIBRAW_SUCCESS){
                DEBUG_p("cannot open %s: %s",filename,libraw_strerror(ret));
                return -1;
            }
            return _convert(h,proc);
        },
        .get_image_surface = [](abydos_plugin_handle_t *h,
                                int page,
                                cairo_bool_t *layer_visibility,
                                int valiant,
                                double time,
                                int frame)
        {
            return cairo_surface_reference(h->surface);
        }
    };
    return &p;
}
