/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <ctype.h>
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */

#include "regisdec.h"


typedef struct{
    int rel;
    enum{
        PT_NONE=0,
        PT_COORD,
        PT_INT,
        PT_LONG,
        PT_ANGLE
    }type;
    double val;
}param_t;

typedef struct{
    int command;
    int firstparam;
}level_t;

struct _regis_decoder_t{
    const regis_interface_t *vt;
    void *userdata;

    double x;
    double y;
    double a;
    int i;
    int l;
    level_t level[4];
    int levelcount;
    param_t param[8];
    int paramcount;
    char numbuf[32];
    int buflen;
};


regis_decoder_t *
regis_decoder_new(const regis_interface_t *vt, void *userdata)
{
    regis_decoder_t *dec = (regis_decoder_t*)malloc(sizeof(regis_decoder_t));
    dec->vt = vt;
    dec->userdata = userdata;
    dec->x = 0;
    dec->y = 0;
    dec->levelcount = 0;
    dec->paramcount = 0;
    dec->buflen = 0;
    return dec;
}


void
regis_decoder_free(regis_decoder_t *dec)
{
    free(dec);
}


static void
_update_position(regis_decoder_t *dec)
{
    if(dec->paramcount < 1)
        return;
    if(!dec->param[0].rel)
        dec->x = 0;
    dec->x += dec->param[0].val;

    if(dec->paramcount < 2)
        return;
    if(!dec->param[1].rel)
        dec->y = 0;
    dec->y += dec->param[1].val;
}


void
regis_decoder_data(regis_decoder_t *dec, int ch)
{
    int paramcount;
    if(isdigit(ch) || '.'==ch || '+'==ch || '-'==ch){
        dec->numbuf[dec->buflen++] = ch;
        return;
    }

    if(dec->buflen>0){
        const char *p = dec->numbuf;
        dec->numbuf[dec->buflen] = '\0';
        dec->param[dec->paramcount].rel = '+'==*p || '-'==*p;
        dec->param[dec->paramcount].val = strtod(p, NULL);
        switch(dec->param[dec->paramcount].type){
        case PT_ANGLE:
            dec->a = dec->param[dec->paramcount].val;
            break;
        case PT_INT:
            dec->i = dec->param[dec->paramcount].val;
            break;
        case PT_LONG:
            dec->l = dec->param[dec->paramcount].val;
        default:
            break;
        }
        ++dec->paramcount;
        dec->buflen = 0;
    }else if(','==ch && PT_COORD==dec->param[dec->paramcount].type){
        dec->param[dec->paramcount].rel = 1;
        dec->param[dec->paramcount].val = 0;
        ++dec->paramcount;
    }

    switch(ch){
    case ',':
    case ' ':
        /* just a parameter separator */
        break;
    case 'A':
        DEBUG_p("param %d is an angle",dec->paramcount);
        dec->param[dec->paramcount].type = PT_ANGLE;
        break;
    case 'I':
        dec->param[dec->paramcount].type = PT_INT;
        break;
    case 'L':
        dec->param[dec->paramcount].type = PT_LONG;
        break;
    case '[':
        dec->param[dec->paramcount+0].type = PT_COORD;
        dec->param[dec->paramcount+1].type = PT_COORD;
        /* fall through */
    case '(':
        ++dec->levelcount;
        //dec->level[dec->levelcount].command = '\0';
        dec->level[dec->levelcount].firstparam = dec->paramcount;
        break;
    case ']':
        _update_position(dec);
        /* fall trough */
    case ')':
        paramcount = dec->level[dec->levelcount].firstparam;
        --dec->levelcount;
        switch(dec->level[dec->levelcount].command){
        case 'C':
            DEBUG_p("curve with %d params", dec->paramcount);
            switch(dec->param[0].type){
            case PT_ANGLE:
                break;
            case PT_COORD:
                dec->vt->curve(dec->userdata,dec->x,dec->y,dec->a);
                break;
            default:
                DEBUG_p("unhandled parameter type to curve (%d)",dec->param[0].type);
                break;
            }
            break;
        case 'F':
            dec->vt->fill(dec->userdata);
            break;
        case 'P':
            dec->vt->position(dec->userdata,dec->x,dec->y);
            break;
        case 'V':
            dec->vt->vector(dec->userdata,dec->x,dec->y);
            break;
        case 'W':
            dec->vt->write(dec->userdata,dec->i);
            break;
        default:
            DEBUG_p("unknown command: %c with %d parameters",dec->level[dec->levelcount].command,dec->paramcount);
            break;
        }
        dec->paramcount = paramcount;
        dec->param[dec->paramcount].type = PT_NONE;
        break;
    default:
        dec->level[dec->levelcount].command = ch;
        dec->a = 360;
        break;
    }
}
