/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <math.h>
#include <libwmf/api.h>

#include "debug.h"
#include "wmfcairo.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    wmfAPI *api;
    enum{
        S_NONE,
        S_MEMOPEN,
        S_FILEOPEN
    }state;
    wmfD_Rect size;
    double xscale;
    double yscale;
    void *userdata;
};


static const char **
_wmf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/wmf",
        "image/x-wmf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_wmf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    wmfAPI_Options opt;
    wmfAPI* api;
    wmf_error_t err;
    abydos_plugin_handle_t *h;

    opt.function = wmf_cairo_function;
    if(wmf_api_create(&api,WMF_OPT_FUNCTION,&opt) != wmf_E_None){
        wmf_api_destroy(api);
        return NULL;
    }

    h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->scalable_size = 1;
    info->threadsafe = 1;
    h->api = api;
    h->state = S_NONE;
    return h;
}


static void
_wmf_free(abydos_plugin_handle_t *h)
{
    switch(h->state){
    case S_NONE:
        break;
    case S_MEMOPEN:
        wmf_mem_close(h->api);
        break;
    case S_FILEOPEN:
        wmf_file_close(h->api);
        break;
    }
    wmf_api_destroy(h->api);
    free(h);
}


static int
_scan(abydos_plugin_handle_t *h){
    float width;
    float height;
    if(wmf_scan(h->api, 0, &h->size) != wmf_E_None)
        return -1;
    wmf_size(h->api,&width,&height);
    //h->info->width = ceil(h->size.BR.x-h->size.TL.x);
    //h->info->height = ceil(h->size.BR.y-h->size.TL.y);
    /* FIXME!! how is the correct size obtained? */
    h->info->width = 800;
    h->info->height = ceil(800*height/width);
    h->xscale = (double)h->info->width/(h->size.BR.x-h->size.TL.x);
    h->yscale = (double)h->info->height/(h->size.BR.y-h->size.TL.y);
    DEBUG_p("size: %dx%d / %fx%f",h->info->width,h->info->height,width,height);
    DEBUG_p("offset: %fx%fx",h->size.TL.x,h->size.TL.y);
    return 0;
}


static int
_wmf_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    if(wmf_mem_open(h->api,(unsigned char *)data,len) != wmf_E_None)
        return -1;
    h->state = S_FILEOPEN;
    return _scan(h);
}


static int
_wmf_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    if(wmf_file_open(h->api,filename) != wmf_E_None)
        return -1;
    h->state = S_FILEOPEN;
    return _scan(h);
}


static void
_wmf_render(abydos_plugin_handle_t *h, cairo_t *cr, int page, cairo_bool_t *visible_layers,int prefered_variant, double time, int frame)
{
    cairo_save(cr);
    cairo_scale(cr,h->xscale,h->yscale);
    h->api->device_data = cr;
    wmf_play(h->api, 0, &h->size);
    cairo_restore(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _wmf_get_mime_types,
        .new = _wmf_new,
        .free = _wmf_free,
        .create_from_data = _wmf_create_from_data,
        .create_from_file = _wmf_create_from_file,
        .render = _wmf_render
    };
    return &p;
}
