/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <cairo-svg.h>
#include <stdlib.h>
#include <string.h>

#include "abydos.h"

/**
 * SECTION: abydos-util
 * @short_description: A few conveniency functions.
 * @title: Utility
 * @include: abydos.h
 **/


/**
 * abydos_create_from_file
 * @mime_type: MIME type
 * @filename: file to load the image from
 *
 * Creates a new #abydos_t object and loads an image file. This is identical
 * to abydos_create() followed by abydos_from_file(). If the operation fails
 * you can call abydos_error() to get an error message. Apart fram that
 * nothing more can be done with @ar and it must be destroyed.
 *
 * returns: An #abydos_t, Or NULL on error.
 */
abydos_t *
abydos_create_from_file(const char *mime_type, const char *filename)
{
    abydos_t *ar = abydos_create(mime_type);
    if(!ar)
        return NULL;
    if(abydos_from_file(ar,filename) < 0){
        abydos_destroy(ar);
        return NULL;
    }
    return ar;
}


/**
 * abydos_load
 * @mime_type: MIME type
 * @filename: file to load the image from
 *
 * Loades an image file and returns it as a cairo image surface.
 *
 * returns: A cairo image surface, or NULL on error.
 */
cairo_surface_t *
abydos_load(const char *mime_type, const char *filename)
{
    abydos_t *ar;
    cairo_surface_t *surface;
    if(!strcasecmp(mime_type,"image/png"))
        return cairo_image_surface_create_from_png(filename);
    ar = abydos_create_from_file(mime_type, filename);
    if(!ar)
        return NULL;
    abydos_set_options(ar,ABYDOS_GROW_TO_ASPECT_RATIO);
    surface = abydos_get_image_surface(ar);
    abydos_destroy(ar);
    return surface;
}


/**
 * abydos_write_to_png
 * @ar: an #abydos_t
 * @filename: file to write to
 *
 * Creates a PNG file with the current state of @ar. Since the PNG format
 * only supports a single image, only one variant of the current frame on
 * the current page will be saved. Also any scalable element will be
 * converted to pixels. You may want to call abydos_set_page_default() and
 * abydos_set_frame_default() before calling this if you don't have a
 * particular wish to save a certain page and/or frame.
 */
cairo_status_t
abydos_write_to_png(abydos_t *ar, const char *filename)
{
    cairo_surface_t *surface = abydos_get_image_surface(ar);
    cairo_surface_write_to_png(surface,filename);
    cairo_surface_destroy(surface);
    return 0;
}


/**
 * abydos_write_to_svg
 * @ar: an #abydos_t
 * @filename: file to write to
 *
 * Creates an SVG file with the current state of @ar. Since the SVG format
 * only supports a single image, only one variant of the current frame on
 * the current page will be saved. You may want to call
 * abydos_set_page_default() and  * abydos_set_frame_default() before
 * calling this if you don't have a particular wish to save a certain page
 * and/or frame.
 */
cairo_status_t
abydos_write_to_svg(abydos_t *ar, const char *filename)
{
    abydos_info_t info;
    cairo_surface_t *surface;
    cairo_t *cr;
    abydos_get_info(ar,&info);
    surface = cairo_svg_surface_create(filename,info.width,info.height);
    cr = cairo_create(surface);
    abydos_render(ar,cr);
    cairo_destroy(cr);
    cairo_surface_destroy(surface);
    return 0;
}


/**
 * abydos_can_animate
 * @ar: an #abydos_t
 *
 * Check if animation can be achieved in some way. That is, if it makes
 * sence to call abydos_get_delay() and update after that specified time.
 *
 * returns: True if animation is possible.
 */
cairo_bool_t
abydos_can_animate(abydos_t *ar)
{
    abydos_info_t info;
    if(abydos_get_frame_count(ar) > 1)
        return 1;
    abydos_get_info(ar, &info);
    return info.features & ABYDOS_FEATURE_SCALABLE_TIME ? 1 : 0;
}
