/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef nile_h
#define nile_h

#include <math.h>                       /* sqrt() */
#include <endian.h>
#include <stddef.h>                     /* size_t */
#include <stdint.h>                     /* uint8_t */
#include <sys/types.h>                  /* off_t */


#define NILE_PREMULTIPLIED (1<<29)
typedef enum{
    NILE_COLORSPACE_UNKNOWN = 0,
    NILE_COLORSPACE_RGB = 1,
    NILE_COLORSPACE_CMYK = 2,
    NILE_COLORSPACE_HSV = 3
}nile_colorspace_t;
typedef enum{
    NILE_CHANNELTYPE_UNKNOWN = 0,
    NILE_CHANNELTYPE_INTEGER = 1,
    NILE_CHANNELTYPE_IEEE754 = 2
}nile_channeltype_t;
#define NILE_ENDIANESS     (1<<23)

/*

31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
         \      / \   /    \      / \         / \         / \         / \         / \         /
R     P     CS      C   E     CB       Alpha     Channel 3   Channel 2   Channel 1   Channel 0

R  = Reserved for custom formats
P  = Premultiplied
CS = Color space
C  = Channel type (integer or IEEE-754 floting point)
E  = Endianness (0 = big endian, 1 = little endian)
CB = Bytes per color channel (offset shifts): up to 128 bytes (1024 bits) per channel

*/

#define NILE_FORMAT_rgba8(r,g,b,a) (NILE_COLORSPACE_RGB<<26|NILE_CHANNELTYPE_INTEGER<<24|0<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgb8(r,g,b) NILE_FORMAT_rgba8(r,g,b,15)
#define NILE_FORMAT_rgba16(r,g,b,a) (NILE_COLORSPACE_RGB<<26|NILE_CHANNELTYPE_INTEGER<<24|1<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgb16(r,g,b) NILE_FORMAT_rgba16(r,g,b,15)
#define NILE_FORMAT_rgba32(r,g,b,a) (NILE_COLORSPACE_RGB<<26|NILE_CHANNELTYPE_IEEE754<<24|2<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgb32(r,g,b) NILE_FORMAT_rgba32(r,g,b,15)

#define NILE_GET_COLORSPACE(f) ((nile_colorspace_t)(((f)>>26)&7))
#define NILE_GET_BYTES_PER_CHANNEL(f) (1<<(((f)>>20)&7))
#define NILE_GET_CHANNEL(f,c) (((f)>>((c)*4))&15)
#define NILE_OFFSET_CHANNEL(f,c) (NILE_GET_CHANNEL(f,c)*NILE_GET_BYTES_PER_CHANNEL(f))
#define NILE_OFFSET_RED(f)   NILE_OFFSET_CHANNEL(f,0)
#define NILE_OFFSET_GREEN(f) NILE_OFFSET_CHANNEL(f,1)
#define NILE_OFFSET_BLUE(f)  NILE_OFFSET_CHANNEL(f,2)
#define NILE_OFFSET_ALPHA(f) NILE_OFFSET_CHANNEL(f,4)

#define NILE_HAS_CHANNEL(f,c) (NILE_GET_CHANNEL(f,c)<15)
#define NILE_HAS_ALPHA(f) NILE_HAS_CHANNEL(f,4)

#define NILE_MAX_BYTES_PER_COLOR 64
#define NILE_STRIDE_ALIGN8  -8
#define NILE_STRIDE_ALIGN16 -16
#define NILE_STRIDE_ALIGN32 -32
#define NILE_STRIDE_ALIGN64 -64
#define NILE_FLIPPED       (1<<0)


typedef enum{
    NILE_TYPE_NONE = 0,
    NILE_TYPE_BITPLANES,
    NILE_TYPE_INDEXED,
    NILE_TYPE_DIRECT
}nile_type_t;

typedef enum{
    NILE_FORMAT_UNKNOWN = 0,
    /* RGB */
    NILE_FORMAT_r8g8b8       = NILE_FORMAT_rgb8(0,1,2),
    NILE_FORMAT_x8r8g8b8     = NILE_FORMAT_rgb8(1,2,3),
    NILE_FORMAT_r8g8b8a8     = NILE_FORMAT_rgba8(0,1,2,3),
    NILE_FORMAT_a8r8g8b8     = NILE_FORMAT_rgba8(1,2,3,0),
    NILE_FORMAT_r16g16b16    = NILE_FORMAT_rgb16(0,1,3),
    NILE_FORMAT_r16g16b16a16 = NILE_FORMAT_rgba16(0,1,2,3),
    NILE_FORMAT_r32g32b32    = NILE_FORMAT_rgb32(0,1,2),
    NILE_FORMAT_r32g32b32a32 = NILE_FORMAT_rgba32(0,1,2,3),
    NILE_FORMAT_r5g6b5       = -10565,
    NILE_FORMAT_r6g6b6       = -10666,
    /* BGR */
    NILE_FORMAT_b8g8r8       = NILE_FORMAT_rgb8(2,1,0),
    NILE_FORMAT_x8b8g8r8     = NILE_FORMAT_rgb8(3,2,1),
    NILE_FORMAT_b8g8r8a8     = NILE_FORMAT_rgba8(2,1,0,3),
    NILE_FORMAT_a8b8g8r8     = NILE_FORMAT_rgba8(3,2,1,0),
    NILE_FORMAT_b16g16r16a16 = NILE_FORMAT_rgba16(2,1,0,3),
    NILE_FORMAT_b32g32r32    = NILE_FORMAT_rgb32(2,1,0),
    NILE_FORMAT_b32g32r32a32 = NILE_FORMAT_rgba32(2,1,0,3),
    /* gray */
    NILE_FORMAT_g8           = NILE_FORMAT_rgb8(0,0,0),
    NILE_FORMAT_g8a8         = NILE_FORMAT_rgba8(0,0,0,1),
    NILE_FORMAT_a8g8         = NILE_FORMAT_rgba8(1,1,1,0),
    NILE_FORMAT_g16          = NILE_FORMAT_rgb16(0,0,0),
    NILE_FORMAT_g16a16       = NILE_FORMAT_rgba16(0,0,0,1),
    NILE_FORMAT_a16g16       = NILE_FORMAT_rgba16(1,1,1,0),
    NILE_FORMAT_g32          = NILE_FORMAT_rgb32(0,0,0)
}nile_format_t;

typedef struct{
#if 1
    nile_type_t type;
    size_t pixels_allocation;           /** Amount of memory allocated for the pixels. */
    union{
        uint8_t *pixels;
        uint8_t **bitplanes;
    };
    int width;
    int height;
    int stride;
    int depth;                          /** The number of bitplanes. */
    size_t palette_allocation;          /** Amount of memory allocated for the palette. */
    uint8_t *palette;
    int colors;
    int bpc;
    nile_format_t format;
#else
    int width;
    int height;
    nile_format_t format;
    int bpc;                            /** Bytes per color. */
    int colors;
    int planes;
    uint8_t *palette;
    uint8_t **bitplanes;
    uint8_t *indexed;
    uint8_t *planar;
#endif
}nile_t;

//typedef int(*nile_convert_colors_func_t)(void *userdata,uint8_t *dst,const uint8_t *src,size_t count);

nile_t *nile_init(nile_t*);
void nile_done(nile_t*);
nile_t *nile_new(void);
void nile_free(nile_t*);

const uint8_t *nile_get_color(const nile_t*, int x, int y);

void nile_drop_pixels(nile_t*);
int nile_ensure_bitplanes(nile_t *,int width, int height, int depth, int deep);
int nile_ensure_indexed(nile_t *,int width, int height);
int nile_ensure_direct(nile_t *,int width, int height, int bytes_per_color);

void nile_drop_palette(nile_t*);
int nile_ensure_palette(nile_t *,int colors, int bytes_per_color);

int nile_get_stride(const nile_t*);
int nile_get_skip(const nile_t*);

const uint8_t *nile_direct_from_indexed(nile_t*);

double nile_color_variance(const uint8_t *color1, nile_format_t format1, const uint8_t *color2, nile_format_t format2);
#define nile_color_distance(c1,f1,c2,f2) sqrt(nile_color_variance(c1,f1,c2,f2))
int nile_best_color_index(const uint8_t *palette, int colors, int bpc, nile_format_t palette_format, const uint8_t *color, nile_format_t color_format);

typedef struct _nile_blit_t nile_blit_t;
typedef void(*nile_blit_func_t)(nile_blit_t*,uint8_t *dst,const uint8_t *src,int count);
typedef void(*nile_blit_bitplanes_func_t)(nile_blit_t*,uint8_t *dst, uint8_t **src, off_t, int count);
struct _nile_blit_t{
    union{
        nile_blit_func_t blit_func;
        nile_blit_bitplanes_func_t blit_bitplanes_func;
    };
    uint8_t tmp_palette[NILE_MAX_BYTES_PER_COLOR * 256];
    const uint8_t *palette;
    int dst_bpc;
    union{
        int src_bpc;
        int src_depth;
    };
    int dst_o[5];
    int src_o[5];
};
nile_blit_t *nile_blit_init_bitplanes_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format);
nile_blit_t *nile_blit_init_packed_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format);
nile_blit_t *nile_blit_init_indexed_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format);
nile_blit_t *nile_blit_init_direct_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int src_bpc, nile_format_t src_format);
//nile_blit_t *nile_blit_init_direct_from_bitplanes_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int depth,
#define nile_blit_line(a,b,c,d) (a)->blit_func(a,b,c,d)
#define nile_blit_bitplanes_line(a,b,c,d,e) (a)->blit_bitplanes_func(a,b,c,d,e);
void nile_blit(nile_blit_t*, uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int width, int height);
void nile_blit_bitplanes(nile_blit_t*, uint8_t *dst, int dst_stride, const uint8_t **src, int src_stride, int depth, int width, int height);
void nile_blit_bitplanes_from_indexed(uint8_t **dst, int dst_stride, int depth, const uint8_t *src, int src_stride, int width, int height);
void nile_blit_indexed(uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int width, int height, int flags);
void nile_blit_indexed_from_bitplanes(uint8_t *dst, int dst_stride, uint8_t **src, int src_stride, int depth, int width, int height);
void nile_blit_direct_rgb(uint8_t *dst, int dst_stride, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_stride, int src_bpc, nile_format_t src_format, int width, int height, int flags);
void nile_blit_direct_from_bitplanes_rgb(uint8_t *dst, int dst_stride, int dst_bpc, nile_format_t dst_format, uint8_t **src, int src_stride, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format, int width, int height, int flags);

void nile_blit_direct_from_indexed_rgb(uint8_t *dst, int dst_stride, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_stride, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format, int width, int height, int flags);

void nile_copy_colors_rgb(uint8_t *dst, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_bpc, nile_format_t src_format, int count);

void nile_copy(nile_t*,const nile_t*,int deep);
int nile_copy_bitplanes_from_indexed(nile_t*,const nile_t*,int deep);
int nile_copy_indexed_from_bitplanes(nile_t*,const nile_t*,int deep);
int nile_copy_direct_from_bitplanes(nile_t*,const nile_t*);
int nile_copy_direct_from_indexed(nile_t*,const nile_t*);
void nile_move(nile_t*,nile_t*);
void nile_copy_pixels(nile_t*,const nile_t*,int deep);
int nile_copy_pixels_bitplanes_from_indexed(nile_t*,const nile_t*);
int nile_copy_pixels_indexed_from_bitplanes(nile_t*,const nile_t*);
void nile_move_pixels(nile_t*,nile_t*);
void nile_copy_palette(nile_t*,const nile_t*,int deep);
void nile_move_palette(nile_t*,nile_t*);

#endif /* nile_h */
